/**************************************************************************
*
*     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
*     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
*     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
*     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
*     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
*     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
*     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
*     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
*     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
*     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
*     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
*     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*     FOR A PARTICULAR PURPOSE.
*     
*     (c) Copyright 2008 Xilinx, Inc.
*     All rights reserved.
*
**************************************************************************/

/**************************************************************************
* Filename:     TestApp_Crash.c
* 
* Description:    
* This application will generate PPC440 exceptions by intentionally 
* performing various software errors.
* 
*
* MODIFICATION HISTORY:
*
* Ver    Who  Date     Changes
* ------ ---- -------- -------------------------------------------------------
* 1.00.a bh   3/30/08 Initial release
*
*
*
******************************************************************************/



/***************************** Include Files *********************************/
#include "xparameters.h"

#include "stdio.h"

#include "xutil.h"
#include "xuartns550_l.h"
#include "xpseudo_asm_gcc.h"
#include "xtlb_l.h"
#include "xexception_l.h"

#include <ppc-asm.h>


/*****************************************************************************/

#define UNMAPPED_INSTRUCTION_ADDRESS   0xB0000000
#define NO_ACCESS_PERMS_ADDRESS        0xB0000004
#define NO_PERIPHERAL_PRESENT_ADDRESS  0xB0000008
#define UNMAPPED_DATA_ADDRESS          0xB000000C

#define TS_INSTRUCTION 0
#define TS_DATA        1

void crash_function (void);
unsigned crash_instructions[10];

/*
 * gen_illegal_instr:
 * This function will generate an Illegal Instruction Program Exception
 * by filling the array crash_instructions[] with
 * arbitrary data, then branching to this array address causing the PPC
 * to attempt to execute this data as code.
 */
void
gen_illegal_instr (void) {
    void (*fp) (void);

    xil_printf("Entering gen_illegal_instr\n\r");

    /*
     * Fill crash_instructions with arbitrary data.
     */
    crash_instructions[0] = 0;
    crash_instructions[1] = 1;
    crash_instructions[2] = 2;
    crash_instructions[3] = 3;
    /*
     * Branch to address of the array crash_instructions[]
     * (which contains only data, not valid instructions)
     */
    fp = (void*)crash_instructions;
    xil_printf("Branching to 0x%08x\n\r", (unsigned)fp);
    (*fp)();

    xil_printf("Exiting gen_illegal_instr\n\r");
}

/*
 * gen_machine_check:
 * This function will generate a Machine Check (if Machine Check
 * has been enabled).  It accomplishes this by reading from an address
 * which does not belong to any peripheral in the system (resulting in
 * a bus timeout).
 */
void gen_machine_check (void) {
    volatile unsigned *ptr;
    unsigned val;

    xil_printf("Entering gen_machine_check\n\r");

    ptr = (unsigned*) NO_PERIPHERAL_PRESENT_ADDRESS;
    xil_printf("Reading from 0x%x\n\r", (unsigned)ptr);
    val = *ptr;
    xil_printf("Read 0x%x\n\r", val);
   
    xil_printf("Exiting gen_machine_check\n\r");
}

/*
 * gen_instr_tlb_miss:
 * This function will generate a TLB Miss (instruction) exception by removing
 * the presently executing address space from the TLB.
 */
void gen_instr_tlb_miss (void) {
    void (*fp) (void);
    short page;
   
    xil_printf("Entering gen_instr_tlb_miss\n\r");

    page = XTlb_FindPage(UNMAPPED_INSTRUCTION_ADDRESS, 0, TS_INSTRUCTION);
    xil_printf("Erasing TLB entry #%d\n\r", page);
    XTlb_ErasePage(page);
    /* Without a context synchronization here, a the following will not cause
     * a TLB miss.
     */
    XTlb_Commit(); 

    /*
     * Branch to address 0xB0000000 
     * (which no longer has a TLB mapping)
     */
    fp = (void*)UNMAPPED_INSTRUCTION_ADDRESS;
    xil_printf("Branching to 0x%08x\n\r", (unsigned)fp);
    (*fp)();

    xil_printf("Exiting gen_instr_tlb_miss\n\r");
}

/*
 * gen_data_tlb_miss:
 * Cause a TLB Miss (data) exception to be generated.  
 * Unmap a page of memory, and then access that virtual address.
 */
void gen_data_tlb_miss (void) {
    volatile unsigned *ptr;
    unsigned val, page;

    xil_printf("Entering gen_data_tlb_miss\n\r");

    /* The virtual address to access */
    ptr = (unsigned*) UNMAPPED_DATA_ADDRESS;
    
    /*
     * Unmap this page 
     */
    page = XTlb_FindPage(UNMAPPED_DATA_ADDRESS, 0, TS_DATA);
    xil_printf("Erasing TLB entry #%d\n\r", page);
    XTlb_ErasePage(page);
    /* Without a context synchronization here, a the following will not
     * cause the desired TLB miss.
     */
    XTlb_Commit(); 

    xil_printf("Reading from 0x%x\n\r", (unsigned)ptr);
    /*
     * An exception will occur here.
     */
    val = *ptr;
    xil_printf("0x%x read from 0x%x\n\r", val, (unsigned)ptr);

    xil_printf("Exiting gen_data_tlb_miss\n\r");
}

/*
 * gen_data_storage:
 * This function will generate a Data Storage exception.  Access
 * permissions are removed from a region of memory (though the address
 * mapping remains) in the TLB.  A write to an address in that region
 * is then performed, causing an access permissions violation.
 */
void gen_data_storage (void) {
    volatile unsigned *ptr;
    unsigned val, page;

    xil_printf("Entering gen_data_storage\n\r");

    /*
     * The address we shall access
     */
    ptr = (unsigned*) NO_ACCESS_PERMS_ADDRESS;
    
    /*
     * Which TLB entry maps this virtual address?
     */
    page = XTlb_FindPage(NO_ACCESS_PERMS_ADDRESS, 0, TS_DATA);
    xil_printf("Clearing all access control for TLB entry #%d\n\r",
               page);
    /*
     * Clear all user and supervisor access to this page
     */
    XTlb_SetStorAttribAccessCtrl(page, 0);
    /* Without a context synchronization here, a the following will
     * not cause a TLB error.
     */
    XTlb_Commit(); 

    xil_printf("Writing to 0x%x\n\r", (unsigned)ptr);
    *ptr = val;

    xil_printf("Exiting gen_data_storage\n\r");
}

/*
 * gen_alignment:
 * This function will generate an alignment exception (if they have been
 * enabled) by performing a non-word aligned 32-bit read.
 */
void gen_alignment (void) {
    unsigned foo, bar, val;
    unsigned char *ptr;

    xil_printf("Entering gen_alignment\n\r");

    /*
     * Fill 2 words in memory with this known pattern.
     */
    foo = 0x11223344;
    bar = 0x55667788;
    ptr = (unsigned char *)&foo;
    xil_printf("Reading from 0x%08x\n\r", (unsigned)ptr);
    /*
     * Perform an aligned read to the address where "foo" is stored.
     */
    xil_printf(": 0x%x\n\r", *(unsigned*) ptr);
    ptr += 1;

    xil_printf("Reading from 0x%08x\n\r", (unsigned)ptr);
    /*
     * Perform an unaligned 32-bit read.
     */
    val = *(unsigned*) ptr;
    xil_printf(": 0x%x\n\r", val);

    xil_printf("Exiting gen_alignment\n\r");
}

/*
 * This fuction will demonstrate a common software error and provide a
 * predictable result.  The local variable "mylocalarray" will be accessed
 * past its bounds.  Writing past the end of the array will corrupt the
 * stack.  The below example will overwrite the saved Link Register of the
 * calling function (myfunction2) so that when this function returns it will
 * branch to gen_illegal_instr(), not the expected location within myfunction2.
 */
void
myfunction3 (void) {
    unsigned mylocalarray[10];
    int index;

#define CALLBYFRAME 1
#define CALLERFRAME 2
#define PREVFRAMESP 3
#define SAVEDLR     4

    printf("-- Entering myfunction3() --\r\n");

    index = (sizeof(mylocalarray)/sizeof(unsigned)) + SAVEDLR;
    /* the value of index is out of bounds - this will write past the end
     * of the array.
     */
    mylocalarray[index] = (unsigned) gen_illegal_instr;

    printf("-- Exiting myfunction3() --\r\n");
}

void
myfunction2 (void) {
    printf("-- Entering myfunction2() --\r\n");

    myfunction3();

    printf("-- Exiting myfunction2() --\r\n");
}

void
myfunction1 (void) {
    printf("-- Entering myfunction1() --\r\n");

    myfunction2();

    printf("-- Exiting myfunction1() --\r\n");
}

extern void setup_crashtrace();
extern void enable_unaligned_exception();
extern char inbyte();

/*
 * choose_crash:
 * Prompt the user for which exception should be generated.
 */
void
choose_crash (void)
{
    int c;

    xil_printf("Choose:\n\r");
    xil_printf("1) Illegal Instruction (Program Exception)\n\r");
    xil_printf("2) Machine Check\n\r");
    xil_printf("3) Enable Machine Check Exceptions\n\r");
    xil_printf("4) Alignment\n\r");
    xil_printf("5) Enable Alignment Exceptions\n\r");
    xil_printf("6) Data Storage\n\r");
    xil_printf("7) TLB Miss (data)\n\r");
    xil_printf("8) Stack corruption\n\r");
    xil_printf("9) Setup software exception handler\n\r");
    
    c = inbyte();

    switch(c) {
    case '1': gen_illegal_instr(); break;
    case '2': gen_machine_check(); break;
    case '3': mtspr(XREG_SPR_MCSR, 0xFFC00000); /* clear all */
              XExc_mEnableExceptions(XREG_MSR_MACHINE_CHECK_ENABLE);
              break;
    case '4': gen_alignment(); break;
    case '5': enable_unaligned_exception(); break;
    case '6': gen_data_storage(); break;
    case '7': gen_data_tlb_miss(); break;
    case '8': myfunction1(); break;
    case '9': setup_crashtrace(); break;

    }
}

/*
 * main:
 */
int
main (void)
{

   /* Initialize RS232_Uart - Set baudrate and number of stop bits */
   XUartNs550_SetBaud(XPAR_RS232_UART_1_BASEADDR,
                      XPAR_XUARTNS550_CLOCK_HZ, 9600);
   XUartNs550_mSetLineControlReg(XPAR_RS232_UART_1_BASEADDR,
                                 XUN_LCR_8_DATA_BITS);
   print("-- Entering main() --\r\n");

   while (1) { 
      choose_crash();
   }

   print("-- Exiting main() --\r\n");
   return 0;
}

