/**************************************************************************
*
*     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
*     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
*     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
*     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
*     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
*     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
*     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
*     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
*     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
*     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
*     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
*     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*     FOR A PARTICULAR PURPOSE.
*     
*     (c) Copyright 2008 Xilinx, Inc.
*     All rights reserved.
*
**************************************************************************/

/**************************************************************************
* Filename:     ppc440_crashtrace.c
* 
* Description:    
* setup_crashtrace() will install a handler for various PPC exceptions
* When a software "crash" occurs a stack backtrace will be sent out
* the serial port.  This text can be used as input to the stackscan.tcl script.
* 
*
* MODIFICATION HISTORY:
*
* Ver    Who  Date     Changes
* -----  ---- -------- -------------------------------------------------------
* 1.00.a bh   03/03/08 First release
* 1.01.a bh   06/19/08 Handle TLB exceptions
*
* $Id: ppc440_crashtrace.c,v 1.3 2008/06/19 16:57:07 bhill Exp $
******************************************************************************/


/***************************** Include Files *********************************/
// Located in: ppc440_virtex5_0/include/xparameters.h
#include "xparameters.h"

#include "stdio.h"

#include "xutil.h"
#include "xexception_l.h"
#include "xpseudo_asm_gcc.h"

#include <ppc-asm.h>

/************************** Constant Definitions *****************************/

/********************** Exception Syndrome Register **************************/

/* Exception Syndrome Register bit definitions */
#define ESR_MCI   0x80000000
#define ESR_PIL   0x08000000
#define ESR_PPR   0x04000000
#define ESR_PTR   0x02000000
#define ESR_FP    0x01000000
#define ESR_ST    0x00800000
#define ESR_DLK   0x00300000
#define ESR_AP    0x00080000
#define ESR_PU0   0x00040000
#define ESR_BO    0x00020000
#define ESR_PIE   0x00010000
#define ESR_PCRE  0x00000010
#define ESR_PCMP  0x00000008

enum ESR_VALS {
    EX_MCI = 0,
    EX_PIL,
    EX_PPR,
    EX_PTR,
    EX_FP,
    EX_ST,
    EX_DLK,
    EX_AP,
    EX_PU0,
    EX_BO,
    EX_PIE,
    EX_PCRE,
    EX_PCMP,
    EX_UNKNOWN
};

/* Plain English for ESR fields */
static const char const *esr_strings[] = {
    "Machine Check - Instruction",
    "Illegal Instruction",
    "Privileged Instruction",
    "Trap Instruction",
    "Floating Point Exception",
    "Store Instruction",
    "Locking Exception",
    "AuxProcessor Operation",
    "Unimplemented Operation",
    "Byte Ordering",
    "Imprecise Exception",
    "Condition Register Enable",
    "Floating Point Compare",
    "Program - UNKNOWN"
};

/********************* Machine Check Status Register  ************************/
/* Machine Check Status Register bit definitions */
#define MCSR_MCS  (1 << 31)
#define MCSR_IB   (1 << 30)
#define MCSR_DRB  (1 << 29)
#define MCSR_DWB  (1 << 28)
#define MCSR_TLBP (1 << 27)
#define MCSR_ICP  (1 << 26)
#define MCSR_DCSP (1 << 25)
#define MCSR_DCFP (1 << 24)
#define MCSR_IMPE (1 << 23)

enum MCSR_VALS {
    MCEX_MCS = 0,
    MCEX_IB,
    MCEX_DRB,
    MCEX_DWB,
    MCEX_TLBP,
    MCEX_ICP,
    MCEX_DCSP,
    MCEX_DCFP,
    MCEX_IMPE
};

static const char const *machine_check_exceptions[] = {
    "Summary",
    "Instruction",
    "Data Read",
    "Data Write",
    "TLB Parity",
    "ICache Parity",
    "DCache Search Parity",
    "DCache Flush Parity",
    "Imprecise",
    "UNKNOWN"
};

/************************** Function Prototypes  *****************************/
u32 check_stack_redzone(int verbose);

/*
 * Symbols from the executable image. These are used to try and verify the
 * validity of stack pointers.
 */
extern u32 _stack_end;
extern u32 __stack;

/*
 * sp_val_valid:
 * Does sp_p appear to be a proper memory address for something
 * on the stack?
 */
static int
sp_val_valid (u32 *sp_p)
{
    u32 sp_val;

    sp_val = (u32) sp_p;
    if (((sp_val & 1) == 0) &&           /* word aligned */
        (sp_val >= (u32)&_stack_end) && /* within the stack */
        (sp_val <= (u32)&__stack)) {
        return XTRUE;
    }
    return XFALSE;
}

/*
 * decode_exception_name:
 * For the indicated exception vector (xexc_id) provide a descriptive string.
 */
static u32
decode_exception_name (u32 xexc_id,
                       const char **vector,
                       const char **cause,
                       u32 *display_dear_val)
{
    u32 esr_val, srr_val, op_store;
    u32 mcsr_val;

    *vector = "UNKNOWN";
    *cause = "UNKNOWN";

    srr_val = mfspr(XREG_SPR_SRR0);

    esr_val = mfspr(XREG_SPR_ESR);
    if (esr_val & ESR_ST) {
        op_store = XTRUE;
    } else {
        op_store = XFALSE;
    }

    /* Come up with a string to display for this type of exception.
     */
    switch (xexc_id) {
    case XEXC_ID_PROGRAM_INT:
        *vector = "Program";

        *cause = esr_strings[EX_UNKNOWN];
        if (esr_val & ESR_PIL) {
            *cause = esr_strings[EX_PIL];
        } else if (esr_val & ESR_PPR) {
            *cause = esr_strings[EX_PPR];
        } else if (esr_val & ESR_PTR) {
            *cause = esr_strings[EX_PTR];
        } else if (esr_val & ESR_PU0) {
            *cause = esr_strings[EX_PU0];
        } else if (esr_val & ESR_PIE) {
            *cause = esr_strings[EX_PIE];
        } else if (esr_val & ESR_PCRE) {
            *cause = esr_strings[EX_PCRE];
        } else if (esr_val & ESR_PCMP) {
            *cause = esr_strings[EX_PCMP];
        }
        break;
    case XEXC_ID_DATA_TLB_MISS_INT:
        *vector = "TLB miss (data)";

        if (op_store) { 
            *cause = "store";
        } else {
            *cause = "load";
        }
        *display_dear_val = XTRUE;
        break;
    case XEXC_ID_INSTRUCTION_TLB_MISS_INT:
        *vector = "TLB MISS (instruction)";
        *cause = NULL;
        break;
    case XEXC_ID_DATA_STORAGE_INT:
        *vector = "Data Storage";

        if (esr_val & ESR_DLK) {
            *cause = esr_strings[EX_DLK];
        } else if (esr_val & ESR_BO) {
            if (op_store) {
                *cause = "Byte Ordering - Store Instruction";
            } else {
                *cause = "Byte Ordering - Load Instruction";
            }
        } else {
            if (op_store) {
                *cause = "Access - Store Instruction";
            } else {
                *cause = "Access - Load Instruction";
            }
        }
        *display_dear_val = XTRUE;
        break;
    case XEXC_ID_ALIGNMENT_INT:
        *vector = "Alignment";
        if (op_store) {
            *cause = "store";
        } else {
            *cause = "load";
        }
        *display_dear_val = XTRUE;
        break;
    case XEXC_ID_MACHINE_CHECK:

        *vector = "Machine Check";

        mcsr_val = mfspr(XREG_SPR_MCSR);
        if (mcsr_val & MCSR_IB) {
            *cause = machine_check_exceptions[MCEX_IB];
        } else if (mcsr_val & MCSR_DRB) {
            *cause = machine_check_exceptions[MCEX_DRB];
        } else if (mcsr_val & MCSR_DWB) {
            *cause = machine_check_exceptions[MCEX_DWB];
        } else if (mcsr_val & MCSR_TLBP) {
            *cause = machine_check_exceptions[MCEX_TLBP];
        } else if (mcsr_val & MCSR_ICP) {
            *cause = machine_check_exceptions[MCEX_ICP];
        } else if (mcsr_val & MCSR_DCSP) {
            *cause = machine_check_exceptions[MCEX_DCSP];
        } else if (mcsr_val & MCSR_DCFP) {
            *cause = machine_check_exceptions[MCEX_DCFP];
        } else if (mcsr_val & MCSR_IMPE) {
            *cause = machine_check_exceptions[MCEX_IMPE];
        }

        srr_val = mfspr(XREG_SPR_MCSRR0);
        break;
    }

    return srr_val;
}

/* 
 * ProgramExceptionHandler:  called by the exception vector handler.  
 * This function prints some debugging information when a exception
 * occurs.
 */
static void
ProgramExceptionHandler (unsigned *ptr)
{
    u32 srr_val, dear_val, display_dear_val, *sp_val;
    u32 xexc_id;
    const char *vector;
    const char *cause;
    static int handling_exception = XFALSE;

    xexc_id = (u32) ptr;

    /* 
     * Did we suffer a second exception in this function ?
     */
    if (handling_exception) {
        xil_printf("\n\rRecursive exceptions.\n\r");
        /*
         * Crash in this exception handler. Do not proceed.
         * Wait here until reset.
         */
        while (1) { ; }
    }
    handling_exception = XTRUE;

    /*
     * Get the SRR0, ESR, and R1 (stack ptr) register values.
     */
    sp_val = (u32*) mfgpr(1);
    dear_val = mfspr(XREG_SPR_DEAR);
    display_dear_val = XFALSE;

    srr_val = decode_exception_name(xexc_id,
                                    &vector,
                                    &cause,
                                    &display_dear_val);
   
    xil_printf("\n\r\n\r");
    /*
     * Display maximum stack depth.
     */
    check_stack_redzone(XTRUE);
    xil_printf("=========================================================\n\r");
    if (display_dear_val) {
        xil_printf("Data Exception Address: 0x%08x\n\r", dear_val);
    }
    if (cause) {
        xil_printf("%s - %s Exception at: 0x%08x\n\r", vector, cause, srr_val); 
    } else {
        xil_printf("%s Exception at: 0x%08x\n\r", vector, srr_val); 
    }

    /* Skip the current stack frame */
    sp_val = (u32*) *sp_val;

    /*
     * _vectorXXX (non_critical_interrupt, critical_interrupt, machine_check)
     * The link register at exception time is saved on this stack frame. It
     * is not in the usual location.
     * See libsrc/standalone_v2_00_a/src/xvectors.S
     */
    /* Saved LR */
    xil_printf("FRAME: 0x%08x RETURN 0x%08x\n\r", (unsigned)sp_val,
                                                  *(sp_val + 4));

    /* Leave Exception Vector frame */
    sp_val = (u32*) *sp_val;

    /*
     * Walk the stack, printing call linkage information.
     */
    while (sp_val_valid(sp_val)) { 
        xil_printf("FRAME: 0x%08x RETURN 0x%08x\n\r", (unsigned)sp_val,
                                                  *(sp_val + 1));
        sp_val = (u32*) *sp_val;
    }
    xil_printf("=========================================================\n\r");

    handling_exception = XFALSE;

    /* Wait here until reset */
    while (1) { ; }
}

/*
 * Pattern used to fill empty stack space.
 */
#define STACK_REDZONE_PATTERN 0xFEEDBEEF

/*
 * check_stack_redzone:
 * Determine maximum stack depth by looking for what remains of the pattern
 * the stack was previously filled with.
 */
u32
check_stack_redzone (int verbose)
{
    u32 *sp_val, stack_total, stack_available;

    sp_val = &_stack_end;
    sp_val++;

    /*
     * Start at the bottom of the stack, moving upwards until the pattern
     * is no longer found.
     */
    while (sp_val < &__stack) {
        if (*sp_val != STACK_REDZONE_PATTERN) {
            break;
        }
        sp_val++;
    }

    /*
     * How much of the stack has never been touched?
     */
    stack_available = (u32)sp_val - (u32)&_stack_end;

    if (verbose) {
        stack_total = (u32)&__stack - (u32)&_stack_end;
        xil_printf("STACK MAXIMUM DEPTH: 0x%x bytes out of 0x%x total.\n\r",
                   (stack_total - stack_available),
                   stack_total);
    }

    /*
     * return space presently available.
     */
    return stack_available;
}

/*
 * create_stack_redzone:
 * Fill in presently unused stack space with a known pattern.  At a later time
 * the stack can be checked against this known pattern to determine a maximim
 * stack usage.  This is useful for determining if sufficient stack space is
 * available.
 */
static void
create_stack_redzone (void)
{
    u32 *r1_val;

    r1_val = (u32*)mfgpr(1);
    r1_val--;

    while (r1_val > &_stack_end) {
        *r1_val = STACK_REDZONE_PATTERN;
        r1_val--;
    }
}

/*
 * enable_unaligned_exception:
 */
void
enable_unaligned_exception (void)
{
    u32 ccr0_val;

    /*
     * Enable exceptions for unaligned word accesses.
     */
    ccr0_val = mfspr(XREG_SPR_CCR0);
    ccr0_val |= XREG_CCR0_FORCE_LOAD_STORE_ALIGN;
    mtspr(XREG_SPR_CCR0, ccr0_val);
    sync;
    isync;
}

/*
 * install exception vectors.  Add a handler for the PROGRAM exception to
 * display debug information when software "crashes".
 */
void
setup_crashtrace (void) 
{
    xil_printf("Installing exception handlers.\n\r");

    /*
     * Reset does not initilize much.  Be certain that no stale error
     * status remains.
     */
    mtspr(XREG_SPR_ESR, 0);
    mtspr(XREG_SPR_MCSR, 0);

    /*
     * Fill presently unused stack space with a known pattern.
     */
    create_stack_redzone();

    /*
     * Initialize the PPC exception table
     */
    XExc_Init();

    /*
     * Register the Program Exception Handler with the exception table
     */
    XExc_RegisterHandler(XEXC_ID_PROGRAM_INT,
                         (XExceptionHandler) ProgramExceptionHandler,
                         (void*)XEXC_ID_PROGRAM_INT);
    XExc_RegisterHandler(XEXC_ID_DATA_TLB_MISS_INT,
                         (XExceptionHandler) ProgramExceptionHandler,
                         (void*)XEXC_ID_DATA_TLB_MISS_INT);
    XExc_RegisterHandler(XEXC_ID_INSTRUCTION_TLB_MISS_INT,
                         (XExceptionHandler) ProgramExceptionHandler,
                         (void*)XEXC_ID_INSTRUCTION_TLB_MISS_INT);
    XExc_RegisterHandler(XEXC_ID_DATA_STORAGE_INT,
                         (XExceptionHandler) ProgramExceptionHandler,
                         (void*)XEXC_ID_DATA_STORAGE_INT);

    XExc_RegisterHandler(XEXC_ID_ALIGNMENT_INT,
                         (XExceptionHandler) ProgramExceptionHandler,
                         (void*)XEXC_ID_ALIGNMENT_INT);
#ifdef NOTNOW
    /*
     * Enable exceptions for unaligned word accesses.
     */
    enable_unaligned_exception();
#endif

    XExc_RegisterHandler(XEXC_ID_MACHINE_CHECK,
                         (XExceptionHandler) ProgramExceptionHandler,
                         (void*)XEXC_ID_MACHINE_CHECK);
#ifdef NOTNOW
    /* 
     * Enable Machine Check exceptions.
     */ 
    XExc_mEnableExceptions(XREG_MSR_MACHINE_CHECK_ENABLE);
#endif

}
