/**************************************************************************
*
*     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
*     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
*     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
*     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
*     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
*     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
*     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
*     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
*     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
*     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
*     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
*     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*     FOR A PARTICULAR PURPOSE.
*     
*     (c) Copyright 2007 Xilinx, Inc.
*     All rights reserved.
*
**************************************************************************/

/**************************************************************************
* Filename:     ppc_debug_malloc.c.c
* 
* Description:    
* Instrument malloc() and free() to maintain memory usage statistics per
* caller.  Statistics are printed with malloc_log_print()
* 
*
* MODIFICATION HISTORY:
*
* Ver   Who  Date     Changes
* ----- ---- -------- -------------------------------------------------------
*
*
*
* $Id: ppc_debug_malloc.c,v 1.1 2007/11/14 21:50:24 bhill Exp $
******************************************************************************/


/***************************** Include Files *********************************/
#include "stdio.h"

#include "xutil.h"
#include "xuartns550_l.h"

#include "xpseudo_asm_gcc.h"
#include <ppc-asm.h>


/************************** Constant Definitions *****************************/
#define MAX_MALLOC_CALLER_LOG 50  /* Maximium callers of malloc() to trace */

/**************************** Type Definitions *******************************/
/* 
 * Data to be recorded in allocated memory when malloc() is called
 */
typedef struct malloc_extraheader malloc_extraheader_t;
struct malloc_extraheader {
    u32    caller_return;
    Xint32 bytes;
    u32    user_malloced_bytes[0];
};

/* 
 * Data to be recorded in debug array when malloc() is called
 */
typedef struct malloc_data malloc_data_t;
struct malloc_data {
    u32    caller_return; /* Stats per caller PC */
    Xint32 bytes;         /* Bytes allocated from this caller */
    u32    total_allocs;  /* Total times malloc() called from this caller */
    u32    total_frees;   /* number of these allocs which were later freed */
};

/*
 * data from each malloc() and free() are stored here.  This is statically
 * sized since it is memory allocation itself being debugged.
 */
static malloc_data_t malloc_log_data[MAX_MALLOC_CALLER_LOG];

/*
 * Maintain some global malloc statistics
 */
static u32 total_malloc;
static u32 total_free;
static u32 malloc_failures;

/*
 * malloc_log_print:
 * Display accumulated malloc() and free() statistics.
 */
void
malloc_log_print (void)
{
    int index;

    printf("MALLOC Statistics:\n\r");
    printf("Total calls to malloc: %d\n\r", total_malloc);
    printf("Total calls to free:   %d\n\r", total_free);
    printf("Malloc failures:       %d\n\r", malloc_failures);

    for (index = 0; index < MAX_MALLOC_CALLER_LOG; index++) {
        if ((malloc_log_data[index].caller_return) &&
            (malloc_log_data[index].bytes != 0)) {

            printf("MALLOC CALLER: 0x%08x TOTAL BYTES: %8d "
                   "ALLOCS: %5d FREES %5d\n\r", 
                malloc_log_data[index].caller_return,
                malloc_log_data[index].bytes,
                malloc_log_data[index].total_allocs,
                malloc_log_data[index].total_frees);
        }
    }
}

/* 
 * update_malloc_log_alloc:
 * Record this call to malloc()
 */
static void
update_malloc_log_alloc (malloc_extraheader_t *hdr)
{
    int index;

    for (index = 0; index < MAX_MALLOC_CALLER_LOG; index++) {
        if ((malloc_log_data[index].caller_return == 0) ||
            (malloc_log_data[index].caller_return == hdr->caller_return)) {
            malloc_log_data[index].caller_return = hdr->caller_return;
            malloc_log_data[index].bytes += hdr->bytes;
            malloc_log_data[index].total_allocs++;
            return;
        }
    }
}

/*
 * update_malloc_log_free:
 * Record this call to free()
 */
static void
update_malloc_log_free (malloc_extraheader_t *hdr)
{
    int index;

    for (index = 0; index < MAX_MALLOC_CALLER_LOG; index++) {
        /*
         * PC of the caller to malloc() was stored in the private header.
         * Look for a match now.
         */
        if (malloc_log_data[index].caller_return == hdr->caller_return) {
            malloc_log_data[index].bytes -= hdr->bytes;
            malloc_log_data[index].total_frees++;
            return;
        }
    }
 
    /*
     * No call to malloc() found in the malloc_log_data[] array
     */
    printf("ERROR: Attempt to free unallocated memory! (hdr:0x%08x)\n\r",
           (u32)hdr);
}

/*
 * __wrap_malloc:
 * When malloc() is called, linker magic results in this function being called.
 * Debug information is logged and then the real malloc() is called
 */
void *
__wrap_malloc (size_t bytes)
{
    malloc_extraheader_t *allocated;
    u32   lr_val;

    /*
     * Determine where malloc() was called from
     */
    lr_val = mfspr(XREG_SPR_LR); 

    /*
     * Allocate the callers memory as requested, plus a little extra to hold
     * our private debug info.
     */
    allocated = (malloc_extraheader_t*)
                __real_malloc(bytes + sizeof(malloc_extraheader_t));
    if (allocated == NULL) {
        /*
         * Insufficient contiguous memory available to complete request.
         */
        malloc_failures++;
        return NULL;
    }
    
    /*
     * Update statistics.
     */
    allocated->caller_return = lr_val;
    allocated->bytes = bytes;
    total_malloc++;
    update_malloc_log_alloc(allocated);

    return (void*) &(allocated->user_malloced_bytes[0]);
}

/*
 * these items from the symbol table are used to verifty that an address
 * passed to free() belongs on the heap.
 */
extern void * _heap_start;
extern void * _heap_end;

/*
 * __wrap_free:
 * When free() is called, linker magic results in this function being called.
 * Debug information is logged and then the real free() is called
 */
void
__wrap_free (void *ptr)
{
    total_free++;
    /*
     * Debug header preceeds user data.
     */
    ptr -= sizeof(malloc_extraheader_t);
    update_malloc_log_free((malloc_extraheader_t*)ptr);

    /*
     * Verify that this appears to be a valid address.
     */
    if (((u32)ptr < (u32)&_heap_start) ||
        ((u32)ptr > (u32)&_heap_end  )) {
        printf("ERROR: Attempt to free memory not located on the heap! "
               "(hdr:0x%08x)\n\r",
               (u32)ptr);
        return;
    }

    __real_free(ptr);
}
