###############################################################################
##
## Copyright (c) 2007, 2008 Xilinx, Inc.  All rights reserved.
##
## Xilinx, Inc.
## XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
## COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
## ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
## STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
## IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
## FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
## XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
## THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
## ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
## FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
## AND FITNESS FOR A PARTICULAR PURPOSE.
##
## Filename:    marvell_88e1111.tcl
## Author:      Brian Hill
##
## Description:
## Print Marvell 88E1111 PHY registers, decoding interesting configuration
## and status bits
##
## $Id: marvell_88e1111.tcl,v 1.3 2008/04/21 20:48:25 bhill Exp $
##
###############################################################################
#
# EXAMPLE USAGE:
# XMD% marvell_phy_probe 0x81c00000
# M88E1111 Found at PHY 0
#
# XMD% marvell_phy_print 0x81c00000 0
# 0  Control         0x6000 LOOPBACK AUTONEG_DISABLED HALF_DUPLX CFG_100MB
# 1  Status          0x7949 LINK_DOWN
# 2  PHYID0          0x0141
# 3  PHYID1          0x0cc1
# 4  AUTONEG_ADV     0x01e1 100TX_FD 100TX_HD 10TX_FD 10TX_HD
# 5  PARTNER_ABILITY 0x0000
# 6  AUTONEG_EXPAND  0x0004
# 9  1000BT_CONTROL  0x0300 ADV_100BT_FD ADV_100BT_HD
# 10 1000BT_STATUS   0x0000 LR_NOT_OK RR_NOT_OK
# 16 PHY_SPCFC_CTRL  0x0078 MDI_AUTO
# 17 PHY_SPCFC_STAT  0x4940 SPD_100MB HALF_DUPLEX SPD_DPLX_RSLVD MDIX
# 18 INT_ENABLE      0x0000
# 19 INT_STATUS      0x0040
# 20 EXT_PYSPFC_CTRL 0x0c60
# 21 RECV_ERR_CNTR   0x0000
# 24 LED_CONTROL     0x4100
# 

puts "marvell_phy_print"
puts "marvell_phy_probe"

# Read the PHY ID from each valid PHY address until a M88e1111 is found
#
proc marvell_phy_probe {temacbase} {
    # Read register offset definitions
    global env
    source "$env(XMD_USER_TCLDIR)/include/marvell_88e1111_defs.tcl"
    set temac_procs [info procs lltemac_read_phy] 
    if {$temac_procs != "lltemac_read_phy"} {
        source "$env(XMD_USER_TCLDIR)/lltemac.tcl"
    }

    set phyaddr 0
    set reg $MARVELL_PHY_ID0_REG
    while {$phyaddr < 31} {
        set regval [lltemac_read_phy $temacbase $phyaddr $reg]
        if {$regval == $MARVELL_88e1111_PHYID0} {
            puts "M88E1111 Found at PHY $phyaddr"
            return
        }
        incr phyaddr
    }
    puts "M88E1111 Not Found"
}

# Read Marvell PHY registers thru the TEMAC via MII
# Decode interesting Config/Status bits into English
#
proc marvell_phy_print {temacbase phyaddr} {
    # Read register offset definitions
    global env
    source "$env(XMD_USER_TCLDIR)/include/marvell_88e1111_defs.tcl"
    set temac_procs [info procs lltemac_read_phy] 
    if {$temac_procs != "lltemac_read_phy"} {
        source "$env(XMD_USER_TCLDIR)/lltemac.tcl"
    }

    # Verify that there really is a Marvell PHY where the caller told us
    # to find one.
    set reg $MARVELL_PHY_ID0_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    if {$regval != $MARVELL_88e1111_PHYID0} {
        puts "No M88E1111 found at PHY $phyaddr"
        return
    }

    set reg $MARVELL_CONTROL_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts -nonewline "$reg  Control         $regval "
    if {$regval & $MARVELL_CONTROL_LOOPBACK} {
        puts -nonewline "LOOPBACK "
    }
    if {$regval & $MARVELL_CONTROL_AUTONEG} {
        puts -nonewline "AUTONEG_ENABLED "
        set autoneg_enabled 1
    } else {
        puts -nonewline "AUTONEG_DISABLED "
        set autoneg_enabled 0
    }
    puts ""

    if {$autoneg_enabled == 0} {
        if {$regval & $MARVELL_CONTROL_DUPLEX} {
            puts -nonewline "CFG_FULL_DUPLX "
        } else {
            puts -nonewline "HALF_DUPLX "
        }

        set config_speed 0
        if {$regval & $MARVELL_CONTROL_SPD_SEL_LSB} {
            set config_speed [expr $config_speed + 1]
        }
        if {$regval & $MARVELL_CONTROL_SPD_SEL_MSB} {
            set config_speed [expr $config_speed + 2]
        }
        switch $config_speed {
            0 {puts -nonewline "CFG_10MB "}
            1 {puts -nonewline "CFG_100MB "}
            2 {puts -nonewline "CFG_1000MB "}
            3 {puts -nonewline "CFG_SPD_RESVD "}
        }
        puts ""
    }

    set reg $MARVELL_STATUS_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts -nonewline "$reg  Status          $regval "
    if {$regval & $MARVELL_STATUS_LINK} {
        puts -nonewline "LINK_UP "
    } else {
        puts -nonewline "LINK_DOWN "
    }
    puts ""

    set reg $MARVELL_PHY_ID0_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts "$reg  PHYID0          $regval"

    set reg $MARVELL_PHY_ID1_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts "$reg  PHYID1          $regval"

    set reg $MARVELL_AUTONEG_ADV_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts -nonewline "$reg  AUTONEG_ADV     $regval "
    if {$regval & $MARVELL_AUTONEG_ADV_NPG} {
        puts -nonewline "NPG "
    }
    if {$regval & $MARVELL_AUTONEG_ADV_PAUSE} {
        puts -nonewline "PAUSE "
    }
    if {$regval & $MARVELL_AUTONEG_ADV_100TX_FD} {
        puts -nonewline "100TX_FD "
    }
    if {$regval & $MARVELL_AUTONEG_ADV_100TX_HD} {
        puts -nonewline "100TX_HD "
    }
    if {$regval & $MARVELL_AUTONEG_ADV_10TX_FD} {
        puts -nonewline "10TX_FD "
    }
    if {$regval & $MARVELL_AUTONEG_ADV_10TX_HD} {
        puts -nonewline "10TX_HD "
    }
    puts ""

    set reg $MARVELL_PARTNER_ABILITY_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts "$reg  PARTNER_ABILITY $regval"

    set reg $MARVELL_AUTONEG_EXPAN_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts "$reg  AUTONEG_EXPAND  $regval"

    set reg $MARVELL_1000BT_CONTROL_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts -nonewline "$reg  1000BT_CONTROL  $regval "
    if {$regval & $MARVELL_1000BT_CONTROL_ADV_1000BT_FD} {
        puts -nonewline "ADV_100BT_FD "
    }
    if {$regval & $MARVELL_1000BT_CONTROL_ADV_1000BT_HD} {
        puts -nonewline "ADV_100BT_HD "
    }
    puts ""

    set reg $MARVELL_1000BT_STATUS_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts -nonewline "$reg 1000BT_STATUS   $regval "
    if {$regval & $MARVELL_1000BT_STATUS_LR_OK} {
        puts -nonewline "LR_OK "
    } else {
        puts -nonewline "LR_NOT_OK "
    }
    if {$regval & $MARVELL_1000BT_STATUS_RR_OK} {
        puts -nonewline "RR_OK "
    } else {
        puts -nonewline "RR_NOT_OK "
    }
    if {$regval & $MARVELL_1000BT_STATUS_LP_1000BT_FD_CAP} {
        puts -nonewline "LP_1000BT_FD_CAP "
    } 
    if {$regval & $MARVELL_1000BT_STATUS_LP_1000BT_HD_CAP} {
        puts -nonewline "LP_1000BT_HD_CAP "
    } 
    puts ""

    set reg $MARVELL_PHY_SPECIFIC_CONTROL_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts -nonewline "$reg PHY_SPCFC_CTRL  $regval "
    set regval [expr $regval & $MARVELL_PHY_SPECIFIC_CONTROL_MDI_MASK] 
    set regval [expr $regval >> $MARVELL_PHY_SPECIFIC_CONTROL_MDI_SHIFT]
    switch $regval {
        0 {puts -nonewline "MDI_FORCE "}
        1 {puts -nonewline "MDI_MDIX "}
        3 {puts -nonewline "MDI_AUTO "}
    }
    puts ""

    set reg $MARVELL_PHY_SPECIFIC_STATUS_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts -nonewline "$reg PHY_SPCFC_STAT  $regval "
    set speed_status [expr $regval >> $MARVELL_PHY_SPCFC_STAT_SPD_SHIFT]
    switch $speed_status {
        3 {puts -nonewline "SPD_RESERVED "}
        2 {puts -nonewline "SPD_1000MB "}
        1 {puts -nonewline "SPD_100MB "}
        0 {puts -nonewline "SPD_10MB "}
    }
    if {$regval & $MARVELL_PHY_SPCFC_STAT_DUPLEX} {
        puts -nonewline "FULL_DUPLEX "
    } else {
        puts -nonewline "HALF_DUPLEX "
    }
    if {$regval & $MARVELL_PHY_SPCFC_STAT_SPD_DUP_RSLVD} {
        puts -nonewline "SPD_DPLX_RSLVD "
    } else {
        puts -nonewline "SPD_DPLX_NOT_RSLVD "
    }
    if {$regval & $MARVELL_PHY_SPCFC_STAT_MDIX} {
        puts -nonewline "MDIX "
    } else {
        puts -nonewline "MDI "
    }
    puts ""
    
    set reg $MARVELL_INT_ENABLE_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts "$reg INT_ENABLE      $regval "

    set reg $MARVELL_INT_STATUS_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts "$reg INT_STATUS      $regval "

    set reg $MARVELL_EXTENDED_PHY_SPECIFIC_CONTROL_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts "$reg EXT_PYSPFC_CTRL $regval "

    set reg $MARVELL_RECEIVE_ERROR_CNTR_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts "$reg RECV_ERR_CNTR   $regval "

    set reg $MARVELL_LED_CONTROL_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts "$reg LED_CONTROL     $regval "
}

