###############################################################################
##
## Copyright (c) 2008 Xilinx, Inc.  All rights reserved.
##
## Xilinx, Inc.
## XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
## COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
## ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
## STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
## IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
## FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
## XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
## THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
## ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
## FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
## AND FITNESS FOR A PARTICULAR PURPOSE.
##
## Filename:    mb_bt.tcl
## Author:      Brian Hill
##
## Description:
## Prints a "Back Trace" from wherever the machine is presently at.
## The last downloaded ELF file is used with addr2line to include symbolic
## information.  The exeucable can only be found if the user has not changed
## directories within XMD.
##
## $Id: mb_bt.tcl,v 1.3 2008/06/24 22:12:25 bhill Exp $
##
###############################################################################

puts "mb_bt"

# mb_lookupsym:
# Use mb-addr2line to find info about the symol at addr
proc mb_lookupsym {addr} {
    # last downloaded executable
    global elf_file

    # nothing downloaded yet?
    if {! [info exists elf_file]} {
        puts ""
        return;
    }
    # File accessable?
    if {! [file exists $elf_file]} {
        puts ""
        return;
    }

    # exec mb-addr2line
    # -It must be in the path.
    # -The user can not have changed directories since downloading.
    set f1 [open "|mb-addr2line -f -s -e $elf_file $addr" r+]
    gets $f1 func
    gets $f1 file
    close $f1

    # Display function and file name for this address
    puts [format "%-17s %s" $func $file]

    return $func
}

# Does addr appear to be a valid memory address ?
proc ptr_valid {addr} {
    if {($addr != 0) && ($addr & 0x3) == 0} {
        return 1
    } else {
        return 0
    }
}

# mb_find_prev_frame:
# For a given stack frame (R1) and the address of an instruction somewhere
# within the function which built that frame (pc_val), find the preceeding
# stack frame.  This is done by examining the instrctions, starting at pc_val,
# and preceeding bacwards (lower memory addresses), looking for an instruction
# which performs math on the stack pointer (R1).
# When an "addik r1, r1, immediate value" instruction is found, the immediate
# value is added to r1_val to find the previous stack frame.
proc mb_find_prev_frame {pc_val  r1_val} {

    # Microblaze "addik r1, r1, immediate value" instruction.
    set ADDIK_R1_R1      0x30210000
    set ADDIK_INSTR_MASK 0xFFFF0000
    set ADDIK_VAL_MASK   0x0000FFFF

    set code_ptr $pc_val
    while {[ptr_valid $code_ptr]} {
        # found "addik r1, r1, val" ?
        set code [format "0x%08x" [mem_read_word $code_ptr]]
        if {($code & $ADDIK_INSTR_MASK) == $ADDIK_R1_R1} {
            # Find the previous stack frame by adding the immediate value used
            # in the addik instruction to the present stack pointer.
            set framesize [expr $code & $ADDIK_VAL_MASK]

            # TCL treats all integers as 32 bits. Sign extend this 16 bit num.
            set framesize [format "0x%08x" [expr $framesize | 0xffff0000]]

            # The code subtracted this number.  For the inverse, we wish to
            # add it.  Negate the number.
            set framesize [expr -$framesize & 0xffff]

            return [format "0x%08x" [expr $r1_val + $framesize]]
        }

        # previous instruction
        incr code_ptr -4
    }

    # Instruction never found.
    return 0
}

#
# reg_trim:
# Convert a string like
#     pc: c0008e50
# to
# 0xc0008e50
# string trimleft "pc: " (for example) does not seem to always work reliably
proc reg_trim {val} {

    # Find the ":" in the string.
    set idx [string first ":" $val]
    # Skip ": "
    incr idx 2
    # New string with just the end of $val
    set newval [string range $val $idx [string length $val]]
    # Make it look like Hex
    set newval [format "0x%08x" 0x0$newval]

    return $newval
}

# mb_bt:
# Display a MicroBlaze "Backtrace"
proc mb_bt { {max_depth 2} } {

    # PC
    set pcval [reg_trim [rrd pc]]

    puts -nonewline "PC:   $pcval            "
    mb_lookupsym $pcval

    # R15 (link register)
    set r15val [reg_trim [rrd r15]]
    puts -nonewline "R15:  $r15val            "
    mb_lookupsym $r15val

    # R1 (stack pointer)
    set r1val [reg_trim [rrd r1]]

    set idx 0
    set func ""
    # Walk the stack.
    while {[ptr_valid $r1val] && $idx < $max_depth} {

        puts -nonewline "R1:   $r1val "

        # Exception handler frame? R15 is stored at a different location.
        # See libsrc/standalone_v2_00_a/src/hw_exception_handler.S
        if {[regexp {handle_other_ex_tail} $func text1]} {
            set saved_r15_loc [expr $r1val + [expr 4*15]]
        } else {
            set saved_r15_loc  $r1val
        }

        # R15 saved in this stack frame
        set saved_r15 [format "0x%08x" [mem_read_word $saved_r15_loc] ]
        puts -nonewline "$saved_r15 "

        if {! [ptr_valid $saved_r15]} {
            puts ""
            return
        }

        set func [mb_lookupsym $saved_r15]

        # Find previous stack frame
        set r1val [mb_find_prev_frame $pcval $r1val]
        set pcval $saved_r15
   
        incr idx
    } 
}

