###############################################################################
##
## Copyright (c) 2008 Xilinx, Inc.  All rights reserved.
##
## Xilinx, Inc.
## XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
## COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
## ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
## STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
## IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
## FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
## XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
## THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
## ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
## FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
## AND FITNESS FOR A PARTICULAR PURPOSE.
##
## Filename:    national_dp83865.tcl
##
## Description:
## Print National DP83865 PHY registers, decoding interesting configuration
## and status bits
##
## $Id: $
##
###############################################################################
#
# EXAMPLE USAGE:
# XMD% national_phy_probe 0x81c00000
# National DP83865_PHYIDR1 Found at PHY 0
#
# XMD% national_phy_print 0x81c00000 1
# 0  Control         0x4040 LOOPBACK AUTONEG_DISABLED HALF_DUPLX CFG_1000MB
# 1  Status          0x794d LINK_UP
# 2  PHYID0          0x2000
# 3  PHYID1          0x5c7a
# 4  AUTONEG_ADV     0x01e1 100TX_FD 100TX_HD 10TX_FD 10TX_HD
# 5  PARTNER_ABILITY 0x0000
# 6  AUTONEG_EXPAND  0x0004
# 9  1000BT_CONTROL  0x0300 ADV_100BT_FD ADV_100BT_HD
# 10 1000BT_STATUS   0x3000 LR_OK RR_OK
# 17 LINK_AN_REG     0x0454 DEEP_LOOPBACK LINK_UP HALF_DUPLEX SPD_1000MB
# 

puts "national_phy_print"
puts "national_phy_probe"

# Read the PHY ID from each valid PHY address until a DP83865 is found
#
proc national_phy_probe {temacbase} {
    # Read register offset definitions
    global env
    source "$env(XMD_USER_TCLDIR)/include/national_dp83865_defs.tcl"
    set temac_procs [info procs lltemac_read_phy] 
    if {$temac_procs != "lltemac_read_phy"} {
        source "$env(XMD_USER_TCLDIR)/lltemac.tcl"
    }

    set phyaddr 0
    set reg $NATIONAL_PHY_ID0_REG
    while {$phyaddr < 31} {
        set regval [lltemac_read_phy $temacbase $phyaddr $reg]
        if {$regval == $NATIONAL_DP83865_PHYID1} {
            puts "National DP83865 Found at PHY $phyaddr"
            return
        }
        incr phyaddr
    }
    puts "National DP83865 Not Found"
}

# Read National PHY registers thru the TEMAC via MII
# Decode interesting Config/Status bits into English
#
proc national_phy_print {temacbase phyaddr} {
    # Read register offset definitions
    global env
    source "$env(XMD_USER_TCLDIR)/include/national_dp83865_defs.tcl"
    set temac_procs [info procs lltemac_read_phy] 
    if {$temac_procs != "lltemac_read_phy"} {
        source "$env(XMD_USER_TCLDIR)/lltemac.tcl"
    }

    # Verify that there really is a National PHY where the caller told us
    # to find one.
    set reg $NATIONAL_PHY_ID0_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    if {$regval != $NATIONAL_DP83865_PHYID1} {
        puts "No National DP83865 found at PHY $phyaddr"
        return
    }

    set reg $NATIONAL_CONTROL_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts -nonewline "$reg  Control         $regval "
    if {$regval & $NATIONAL_CONTROL_LOOPBACK} {
        puts -nonewline "LOOPBACK "
    }
    if {$regval & $NATIONAL_CONTROL_AUTONEG} {
        puts -nonewline "AUTONEG_ENABLED "
        set autoneg_enabled 1
    } else {
        puts -nonewline "AUTONEG_DISABLED "
        set autoneg_enabled 0
    }

    if {$autoneg_enabled == 0} {
        if {$regval & $NATIONAL_CONTROL_DUPLEX} {
            puts -nonewline "CFG_FULL_DUPLX "
        } else {
            puts -nonewline "HALF_DUPLX "
        }

        set config_speed 0
        if {$regval & $NATIONAL_CONTROL_SPD_SEL_LSB} {
            set config_speed [expr $config_speed + 1]
        }
        if {$regval & $NATIONAL_CONTROL_SPD_SEL_MSB} {
            set config_speed [expr $config_speed + 2]
        }
        switch $config_speed {
            0 {puts -nonewline "CFG_10MB "}
            1 {puts -nonewline "CFG_100MB "}
            2 {puts -nonewline "CFG_1000MB "}
            3 {puts -nonewline "CFG_SPD_RESVD "}
        }
    }
    puts ""

    set reg $NATIONAL_STATUS_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts -nonewline "$reg  Status          $regval "
    if {$regval & $NATIONAL_STATUS_LINK} {
        puts -nonewline "LINK_UP "
    } else {
        puts -nonewline "LINK_DOWN "
    }
    puts ""

    set reg $NATIONAL_PHY_ID0_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts "$reg  PHYID0          $regval"

    set reg $NATIONAL_PHY_ID1_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts "$reg  PHYID1          $regval"

    set reg $NATIONAL_AUTONEG_ADV_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts -nonewline "$reg  AUTONEG_ADV     $regval "
    if {$regval & $NATIONAL_AUTONEG_ADV_NPG} {
        puts -nonewline "NPG "
    }
    if {$regval & $NATIONAL_AUTONEG_ADV_PAUSE} {
        puts -nonewline "PAUSE "
    }
    if {$regval & $NATIONAL_AUTONEG_ADV_100TX_FD} {
        puts -nonewline "100TX_FD "
    }
    if {$regval & $NATIONAL_AUTONEG_ADV_100TX_HD} {
        puts -nonewline "100TX_HD "
    }
    if {$regval & $NATIONAL_AUTONEG_ADV_10TX_FD} {
        puts -nonewline "10TX_FD "
    }
    if {$regval & $NATIONAL_AUTONEG_ADV_10TX_HD} {
        puts -nonewline "10TX_HD "
    }
    puts ""

    set reg $NATIONAL_PARTNER_ABILITY_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts "$reg  PARTNER_ABILITY $regval"

    set reg $NATIONAL_AUTONEG_EXPAN_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts "$reg  AUTONEG_EXPAND  $regval"

    set reg $NATIONAL_1000BT_CONTROL_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts -nonewline "$reg  1000BT_CONTROL  $regval "
    if {$regval & $NATIONAL_1000BT_CONTROL_ADV_1000BT_FD} {
        puts -nonewline "ADV_100BT_FD "
    }
    if {$regval & $NATIONAL_1000BT_CONTROL_ADV_1000BT_HD} {
        puts -nonewline "ADV_100BT_HD "
    }
    puts ""

    set reg $NATIONAL_1000BT_STATUS_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts -nonewline "$reg 1000BT_STATUS   $regval "
    if {$regval & $NATIONAL_1000BT_STATUS_LR_OK} {
        puts -nonewline "LR_OK "
    } else {
        puts -nonewline "LR_NOT_OK "
    }
    if {$regval & $NATIONAL_1000BT_STATUS_RR_OK} {
        puts -nonewline "RR_OK "
    } else {
        puts -nonewline "RR_NOT_OK "
    }
    if {$regval & $NATIONAL_1000BT_STATUS_LP_1000BT_FD_CAP} {
        puts -nonewline "LP_1000BT_FD_CAP "
    } 
    if {$regval & $NATIONAL_1000BT_STATUS_LP_1000BT_HD_CAP} {
        puts -nonewline "LP_1000BT_HD_CAP "
    } 
    puts ""

    set reg $NATIONAL_LINK_AN_REG
    set regval [lltemac_read_phy $temacbase $phyaddr $reg]
    puts -nonewline "$reg LINK_AN_REG     $regval "
    if {$regval & $NATIONAL_LINK_AN_FIFO_ERR} {
        puts -nonewline "FIFO_ERR "
    }
    if {$regval & $NATIONAL_LINK_AN_SHLW_LOOP} {
        puts -nonewline "SHALLOW_LOOPBACK "
    }
    if {$regval & $NATIONAL_LINK_AN_DEEP_LOOP} {
        puts -nonewline "DEEP_LOOPBACK "
    }
    if {$regval & $NATIONAL_LINK_AN_LINK_STS} {
        puts -nonewline "LINK_UP "
    } else {
        puts -nonewline "LINK_DOWN "
    }
    if {$regval & $NATIONAL_LINK_AN_DUPLEX_STS} {
        puts -nonewline "FULL_DUPLEX "
    } else {
        puts -nonewline "HALF_DUPLEX "
    }
    set regval [expr $regval >> $NATIONAL_LINK_AN_SPEED_STS_SHIFT]
    set regval [expr $regval & 0x3]
    switch $regval {
        0 {puts -nonewline "SPD_10MB "}
        1 {puts -nonewline "SPD_100MB "}
        2 {puts -nonewline "SPD_1000MB "}
        3 {puts -nonewline "SPD_RSVD "}
    }
    puts ""

}

