###############################################################################
##
## Copyright (c) 2007, 2008 Xilinx, Inc.  All rights reserved.
##
## Xilinx, Inc.
## XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
## COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
## ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
## STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
## IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
## FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
## XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
## THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
## ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
## FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
## AND FITNESS FOR A PARTICULAR PURPOSE.
##
## Filename:    packet_decode.tcl
## Author:      Brian Hill
##
## Description:
## Decode an Ethernet packet in memory
##
## $Id: packet_decode.tcl,v 1.2 2008/04/21 20:48:25 bhill Exp $
##
###############################################################################

# EXAMPLE:
#
# XMD% en_packet_decode 0x5f9200
# ETHERNET:
# 0x005f9200 -> 0x000d566d
# 0x005f9204 -> 0x769f0605 DST_MACADDR: 0x000d566d769f
# 0x005f9208 -> 0x04030201 SRC_MACADDR: 0x060504030201
# 0x005f920c -> 0x0800     ETHERTYPE: 0x0800 IPv4
# IPv4:
# 0x005f920e -> 0x45000028 IPHL 20 TOTAL_LEN 40
# 0x005f9212 -> 0x00044000 Ident: 4 DONOT_FRAG IP_OFF 0
# 0x005f9216 -> 0x4006b975 TTL 64 IP_PROTO 6:TCP HDR_CSUM 47477
# 0x005f921a -> 0xc0a80002 SRC_IP 192.168.0.2
# 0x005f921e -> 0xc0a80004 DST_IP 192.168.0.4
# TCP:
# 0x005f9222 -> 0x32410843 SRC_PORT 12865 DST_PORT 2115
# 0x005f9226 -> 0x00000000 SEQ
# 0x005f922a -> 0x5beeb84a ACK
# 0x005f922e -> 0x50140000 TCP_HDR_LEN 20 RST ACK WINDOW_SIZE 0
# 0x005f9232 -> 0x81710000 TCP_CSUM 33137
# 
# CALCULATED TCP CHECKSUM: 0xdfbc
# 

# this script requires utils.tcl and csum16.tcl :
# source "$env(XMD_USER_TCLDIR)/utils.tcl"
# source "$env(XMD_USER_TCLDIR)/csum16.tcl"

puts "ipv4_packet_decode"
puts "en_packet_decode"


# print_ip_dotdec
# print an IPv4 address in dotted decimal format
proc print_ip_dotdec {ipaddr} {
    set oct1 [expr [expr $ipaddr >> 24] & 0xFF]
    set oct2 [expr [expr $ipaddr >> 16] & 0xFF]
    set oct3 [expr [expr $ipaddr >>  8] & 0xFF]
    set oct4 [expr [expr $ipaddr >>  0] & 0xFF]

    puts -nonewline "$oct1.$oct2.$oct3.$oct4"
}

# icmp_decode
# decode ICMP
proc icmp_decode {bufaddr} {
    set addr $bufaddr

    puts "ICMP:"

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    set type_val [expr $mval >> 24]
    set csum [expr $mval & 0xFFFF]
    puts -nonewline "$addr -> $mval "
    switch $type_val {
        0  { puts -nonewline "ECHO_REPLY " }
        3  { puts -nonewline "DEST_UNREACHABLE " }
        4  { puts -nonewline "SOURCE_QUENCH " }
        5  { puts -nonewline "REDIRECT " }
        8  { puts -nonewline "ECHO " }
        11  { puts -nonewline "TIME_EXCEEDED " }
        12  { puts -nonewline "PARAMETER_PROBLEM " }
        13  { puts -nonewline "TIMESTAMP " }
        14  { puts -nonewline "TIMESTAMP_REPLY " }
        default { puts -nonewline "TYPE:$type_val" }
    }
    puts "CSUM: $csum "
    incr addr 4

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    set identifier [expr [expr $mval >> 16] & 0xFFFF]
    set seq [expr $mval & 0xFFFF]
    puts "$addr -> $mval IDENTIFIER $identifier SEQUENCE NUM $seq"
    incr addr 4

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    puts "$addr -> $mval "
}

# udp_decode
proc udp_decode {bufaddr} {
    set addr $bufaddr

    puts "UDP:"

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    set src_port [expr $mval >> 16]
    set dst_port [expr $mval & 0xFFFF]
    puts "$addr -> $mval SRC_PORT $src_port DST_PORT $dst_port"
    incr addr 4

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    set udp_length [expr $mval >> 16]
    set udp_csum [expr $mval & 0xFFFF]
    puts "$addr -> $mval LENGTH $udp_length CHECKSUM $udp_csum"
}

# tcp_decode
proc tcp_decode {bufaddr} {
    set addr $bufaddr

    puts "TCP:"

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    set src_port [expr $mval >> 16]
    set dst_port [expr $mval & 0xFFFF]
    puts "$addr -> $mval SRC_PORT $src_port DST_PORT $dst_port"
    incr addr 4

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    puts "$addr -> $mval SEQ"
    incr addr 4

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    puts "$addr -> $mval ACK"
    incr addr 4

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    set tcp_hdr_len [expr [expr $mval >> 28] * 4]
    set tcp_flags [expr [expr $mval >> 16] & 0x3F]
    set window_sz [expr $mval & 0xFFFF]
    puts -nonewline "$addr -> $mval "
    puts -nonewline "TCP_HDR_LEN $tcp_hdr_len "
    if {$tcp_flags & 0x01} {
        puts -nonewline "FIN "
    }
    if {$tcp_flags & 0x02} {
        puts -nonewline "SYN "
    }
    if {$tcp_flags & 0x04} {
        puts -nonewline "RST "
    }
    if {$tcp_flags & 0x08} {
        puts -nonewline "PSH "
    }
    if {$tcp_flags & 0x10} {
        puts -nonewline "ACK "
    }
    if {$tcp_flags & 0x20} {
        puts -nonewline "URG "
    }
    if {$tcp_flags & 0x40} {
        puts -nonewline "ECE "
    }
    if {$tcp_flags & 0x80} {
        puts -nonewline "CWR "
    }
    puts "WINDOW_SIZE $window_sz"
    incr addr 4

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    set tcp_csum [expr $mval >> 16]
    puts -nonewline "$addr -> $mval "
    puts "TCP_CSUM $tcp_csum"
    incr addr 4

    # Are there options ?
    if {$tcp_hdr_len > 20} {
        set addr [format "0x%08x" $addr]
        set mval [format "0x%08x" [rdwd_by_hw $addr] ]
        puts -nonewline "$addr -> $mval "

        set tcp_option [expr $mval >> 24]
        puts -nonewline "TCP_OPTION $tcp_option "
        if {$tcp_option == 2} {
            set tcp_mss [expr $mval & 0xFFFF]
            puts -nonewline "MSS $tcp_mss "
        }
    }
    puts ""
}

# tcp_csum
# Calculate a TCP checksum for the datagram in memory at tcp_addr.
# Note that the entire datagram must be in this one buffer.
proc tcp_csum {src_ip dst_ip tcp_len tcp_addr} {

    set sum 0

    # Calculate the IP pseudo-header

    # SA
    set s1 [expr $src_ip & 0xFFFF]
    set s2 [expr [expr $src_ip >> 16] & 0xFFFF]
    set sum [expr $s1 + $s2]
    set sum [expr [expr $sum >> 16] + $sum]

    # DA
    set s1 [expr $dst_ip & 0xFFFF]
    set s2 [expr [expr $dst_ip >> 16] & 0xFFFF]
    set sum [expr $s1 + $s2 + $sum]

    # protocol 6, TCP
    set sum [expr $sum + 6]

    # TCP lenfth
    set sum [expr $sum + $tcp_len]

    # fold
    set sum [expr [expr $sum >> 16] + $sum]
    set sum [expr $sum & 0xFFFF]
    set sum [format "0x%04x" $sum]
    set pseudo_hdr_sum $sum

    # calculate TCP csum, skipping word 8 (checksum field)
    set sum [csum16 $tcp_addr $tcp_len $sum 8]

    puts "CALCULATED TCP CHECKSUM: $sum (pseudo header sum $pseudo_hdr_sum)"
}

proc ipv4_packet_decode {bufaddr} {
    set addr $bufaddr

    puts "IPv4:"

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    puts -nonewline "$addr -> $mval "
    set ip_hdr_len [expr [expr $mval >> 24] & 0xF]
    set ip_hdr_len [expr $ip_hdr_len * 4]
    set total_len  [expr $mval & 0xFFFF]
    puts "IPHL $ip_hdr_len TOTAL_LEN $total_len"
    incr addr 4

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    set ip_id [expr $mval >> 16]
    set ip_off [expr $mval & 0x1fff]
    puts -nonewline "$addr -> $mval "
    puts -nonewline "Ident: $ip_id "
    if {$mval & 0x4000} {
        puts -nonewline "DONOT_FRAG "
    }
    if {$mval & 0x2000} {
        puts -nonewline "MORE_FRAG "
    }
    puts "IP_OFF $ip_off"
    incr addr 4

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    set ttl [expr $mval >> 24]
    set ip_proto [expr [expr $mval >> 16] & 0xFF]
    set hdr_csum [expr $mval & 0xFFFF]
    puts -nonewline "$addr -> $mval "
    puts -nonewline "TTL $ttl "
    puts -nonewline "IP_PROTO $ip_proto:"
    switch $ip_proto {
        1  { puts -nonewline "ICMP " }
        6  { puts -nonewline "TCP " }
        17 { puts -nonewline "UDP " }
    }
    puts "HDR_CSUM $hdr_csum"
    incr addr 4
 
    set addr [format "0x%08x" $addr]
    set src_ip [format "0x%08x" [rdwd_by_hw $addr] ]
    puts -nonewline "$addr -> $src_ip SRC_IP "
    print_ip_dotdec $src_ip
    puts ""
    incr addr 4
  
    set addr [format "0x%08x" $addr]
    set dst_ip [format "0x%08x" [rdwd_by_hw $addr] ]
    puts -nonewline "$addr -> $dst_ip DST_IP "
    print_ip_dotdec $dst_ip
    puts ""
  
    set addr [expr $bufaddr + $ip_hdr_len]
    switch $ip_proto {
        1  { icmp_decode $addr }
        6  { # decode the TCP header
             tcp_decode $addr
             # Calculate the TCP checksum
             set tcp_len [expr $total_len - $ip_hdr_len]
             tcp_csum $src_ip $dst_ip $tcp_len $addr
           }
        17 { udp_decode $addr }
    }
}

proc en_packet_decode {bufaddr} {

    puts "ETHERNET:"

    set addr $bufaddr

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    puts "$addr -> $mval "
    set mac_addr $mval
    incr addr 0x4

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    puts -nonewline "$addr -> $mval "
    puts -nonewline "DST_MACADDR: $mac_addr"
    set mac_addr [expr [expr $mval & 0xffff0000] >> 16]
    set mac_addr [format "%04x" $mac_addr]
    puts $mac_addr
    set mac_addr [expr $mval & 0xFFFF]
    set mac_addr [format "0x%04x" $mac_addr]
    incr addr 4

    set addr [format "0x%08x" $addr]
    set mval [format "0x%08x" [rdwd_by_hw $addr] ]
    puts -nonewline "$addr -> $mval SRC_MACADDR: $mac_addr"
    set mac_addr [format "%08x" $mval]
    puts $mac_addr
    incr addr 4

    # read type/length
    set addr [format "0x%08x" $addr]
    set ethertype [format "0x%04x" [mem_read_hwd $addr] ]
    puts -nonewline "$addr -> $ethertype     "
    incr addr 2
    if {$ethertype <= 0x5DC} {
        puts "LENGTH: $ethertype"
    } else {
        if {$ethertype} {
            puts -nonewline "ETHERTYPE: $ethertype "
        }
        switch $ethertype {
            0x0800  { puts "IPv4"
                      ipv4_packet_decode $addr
                    }
            0x0806  { puts "IPv4_ARP" }
            default { puts "" }
        }
    }

    puts ""
}

