###############################################################################
##
## Copyright (c) 2008 Xilinx, Inc.  All rights reserved.
##
## Xilinx, Inc.
## XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
## COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
## ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
## STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
## IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
## FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
## XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
## THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
## ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
## FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
## AND FITNESS FOR A PARTICULAR PURPOSE.
##
## Filename:    ppc405_dcache.tcl
## Author:      Brian Hill
##
## Description:
## Print PPC405 Data Cache contents.  XMD has pseudo-memory mapped the
## D-Cache tag and data areas.
##
## $Id: ppc405_dcache.tcl,v 1.1 2008/04/02 18:56:58 bhill Exp $
##
###############################################################################
#
# EXAMPLE USAGE:
#
# Address: calculated with Tag + Line
# D:       cache line is dirty
# L:       way is least-recently-used
#
# XMD% ppc405_dcache_print 0x78004000 0x78000000
# Line Tag        Address  D L Data
# 192A 0x00040000 00041800 *   0000000300000004a0a0d223fcc3a2a8
#                              fcc1052513782840413800e630869b82
# 193A 0x00000000 00001820 * * 00000000000000010000000000000000
#                              00000000000000000000000000000000
# 194A 0x00000000 00001840 * * 00000001000000020000000000000000
#                              00000000000000000000000000000000
# 192B 0x00050000 00051800 *   00000004000000056641e6229083b2ad
#                              4095247692599122952a0ae8e420a382
# 
# XMD% mrd 0x41800 8
#  41800:   00000003
#  41804:   00000004
#  41808:   A0A0D223
#  4180C:   FCC3A2A8
#  41810:   FCC10525
#  41814:   13782840
#  41818:   413800E6
#  4181C:   30869B82
#

puts "ppc405_dcache_print"

# ppc405_dcache_print:
#
proc ppc405_dcache_print {tag data} {
    set line 0
    set linesize 32 

    puts "Line Tag        Address  D L Data"
    # examine all 256 lines for Way A and Way B
    while {$line < 512} {
        # read the tag for this line
        set tagw0 [format "0x%08x" [mem_read_word $tag]]
        set tagvalue [format "0x%08x" [expr $tagw0 & 0xFFFFC000]]
        set tagu0    [format "0x%08x" [expr [expr $tagw0 >> 7] & 1]]
        set tagdrty  [format "0x%08x" [expr [expr $tagw0 >> 5] & 1]]
        set tagvld   [format "0x%08x" [expr [expr $tagw0 >> 4] & 1]]
        set taglru   [format "0x%08x" [expr $tagw0 & 1]]

        # Is this line valid (presently in use) ?
        if {$tagvld} {
            if {$line < 256} {
                # Way A 

                # use the tag + line number to calculate the memory address
                set addr [expr $tagvalue + [expr $line << 5]]
                set addr [format "%08x" $addr]
                puts -nonewline "[expr $line]A $tagvalue $addr "
            } else {
                # Way B

                # use the tag + line number to calculate the memory address
                set addr [expr $tagvalue + [expr [expr $line - 256] << 5]]
                set addr [format "%08x" $addr]
                puts -nonewline "[expr $line - 256]B $tagvalue $addr "
            }

            # is this line dirty ?
            if {$tagdrty} {
                puts -nonewline "* "
            } else {
                puts -nonewline "  "
            }

            # Least Recently Used Way?
            if {$taglru} {
                puts -nonewline "* "
            } else {
                puts -nonewline "  "
            }

            # Print the cache data for this line
            set i 0
            while {$i < $linesize} {
                set dataval [format "%08x" [mem_read_word [expr $data + $i]]]
                puts -nonewline "$dataval"
               
                # Print the data on two lines
                incr i 4
                if {$i == 16} {
                    puts -nonewline "\n                             "
                }
            }
            puts ""
        }

        # Next cache line
        incr tag  $linesize
        incr data $linesize
        incr line
    }
}

