###############################################################################
##
## Copyright (c) 2008 Xilinx, Inc.  All rights reserved.
##
## Xilinx, Inc.
## XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
## COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
## ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
## STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
## IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
## FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
## XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
## THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
## ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
## FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
## AND FITNESS FOR A PARTICULAR PURPOSE.
##
## Filename:    ppc405_tlb.tcl
## Author:      Brian Hill
##
## Description:
## Script to display/decode PPC405 TLB entries via XMD "memory map" of these.
##
## $Id: ppc405_tlb.tcl,v 1.3 2008/04/21 20:48:25 bhill Exp $
##
###############################################################################
#
# EXAMPLE USAGE:
#
# XMD% ppc405_read_tlb 0x70004000
# #  RPN        TAG        SIZE  V E U0 EX WR W I M G
# 56 0x00000000 0xc0003000 16MB  V      EX WR
# 57 0x03800000 0xc395d000 4MB   V      EX WR
# 39 0x03837000 0x0ffe9000 4KB   V         WR
# 52 0x038b2000 0x0ff79000 4KB   V      EX
# ...... deleted ....
# 12 0x038c6000 0x0ff19000 4KB   V      EX
# 48 0x81800000 0xfdffe000 4KB   V         WR   I   G
#  1 0x81c00000 0xc40aa000 4KB   V         WR   I   G
# 49 0x83e01000 0xfdfff000 4KB   V         WR   I   G
#

puts "ppc405_read_tlb"

# decode_tlb_lo
# Print out all of the fields of TLBLO
proc decode_tlb_lo {tlb_data} {

    if {$tlb_data & [expr 1 << 9]} {
        puts -nonewline "EX "
    } else {
        puts -nonewline "   "
    }
    if {$tlb_data & [expr 1 << 8]} {
        puts -nonewline "WR "
    } else {
        puts -nonewline "   "
    }
    if {$tlb_data & [expr 1 << 3]} {
        puts -nonewline "W "
    } else {
        puts -nonewline "  "
    }
    if {$tlb_data & [expr 1 << 2]} {
        puts -nonewline "I "
    } else {
        puts -nonewline "  "
    }
    if {$tlb_data & [expr 1 << 1]} {
        puts -nonewline "M "
    } else {
        puts -nonewline "  "
    }
    if {$tlb_data & [expr 1 << 0]} {
        puts -nonewline "G "
    } else {
        puts -nonewline "  "
    }
}

# decode_tlb_hi
# Print out fields of TLBHI
proc decode_tlb_hi {tlb_tag} {

    set size [expr [expr $tlb_tag >> 7] & 0x7]
    switch $size {
     0 {puts -nonewline "1KB   "}
     1 {puts -nonewline "4KB   "}
     2 {puts -nonewline "16KB  "}
     3 {puts -nonewline "64KB  "}
     4 {puts -nonewline "256KB "}
     5 {puts -nonewline "1MB   "}
     6 {puts -nonewline "4MB   "}
     7 {puts -nonewline "16MB  "}
    }

    if {$tlb_tag & [expr 1 << 6]} {
        puts -nonewline "V "
    } else {
        puts -nonewline "  "
    }

    if {$tlb_tag & [expr 1 << 5]} {
        puts -nonewline "E "
    } else {
        puts -nonewline "  "
    }

    if {$tlb_tag & [expr 1 << 4]} {
        puts -nonewline "U0 "
    } else {
        puts -nonewline "   "
    }
}

# decode_tlb_data_rpn
# Return the RPN field of TLBLO
proc decode_tlb_data_rpn {tlb_data} {
    set rpn [format "0x%08x" [expr $tlb_data & 0xFFFFF800]]
    return $rpn
}

# decode_tlb_tag
# Return the TAG field of TLBHI
proc decode_tlb_tag {tlb_tag} {
    set tag [format "0x%08x" [expr $tlb_tag & 0xFFFFF800]]
    return $tag
}

# tlb_list_insert
# Insert new_item into tlb_list by order of increasing RPN
proc tlb_list_insert {tlb_list new_item} {

    set new_tlb_entry_rpn [lindex $new_item 1]

    set tlb_list_len [llength $tlb_list]
    set item 0
    # Find where new_item belongs. Walk the entire list.
    while {$item < $tlb_list_len} {
        set list_item [lindex $tlb_list $item]
        set tlb_entry_num [lindex $list_item 0]
        set tlb_entry_rpn [lindex $list_item 1]

        # new_item RPN smaller than this list item RPN ?
        if {$new_tlb_entry_rpn <= $tlb_entry_rpn} {
            # Insert new_item
            set tlb_list [linsert $tlb_list $item $new_item]
            break
        }
 
        incr item
    }
    # new_item at the end of list? (Largest item in the list)
    if {$item == $tlb_list_len} {
        # Insert new_item at end of list
        set tlb_list [linsert $tlb_list $item $new_item]
    }

    return $tlb_list
}

# ppc405_read_tlb
# Read XMD "memory mapped" processor TLB entries.
# All valid entries are displayed, sorted by Real Page Number
proc ppc405_read_tlb {base} {
    set addr $base

    puts "#  RPN        TAG        SIZE  V E U0 EX WR W I M G " 

    # create an ordered list of the TLB entries, sorted by RPN
    set tlb_num 0
    while {$tlb_num < 64} {
        set tlb_tag_addr [format "0x%08x" [expr $addr + 4]]
        set tlb_data_addr [format "0x%08x" [expr $addr + 8]]

        set mval [format "0x%08x" [mem_read_word $tlb_tag_addr] ]
        # Valid entry?   
        if {$mval & [expr 1 << 6]} {
            set mval [format "0x%08x" [mem_read_word $tlb_data_addr] ]
            set tlb_rpn [list $tlb_num [decode_tlb_data_rpn $mval]]
            # First one?
            if { [info exists tlb_list] } {
                # Insert this entry into the list
                set tlb_list [tlb_list_insert $tlb_list $tlb_rpn]
            } else {
                # First list item
                set tlb_list [list $tlb_rpn]
            }
        }

        incr addr 16
        set addr [format "0x%08x" $addr]
        incr tlb_num
    }

    # Print all valid TLB entries as present in the ordered list
    set tlb_list_len [llength $tlb_list]
    set idx 0
    while {$idx < $tlb_list_len} {
        # Get TLB entry from the list
        set list_item [lindex $tlb_list $idx]
        # Fetch TLB entry number (0-63) and that entry's RPN from this item
        set tlb_entry_num [format "%2d" [lindex $list_item 0]]
        set tlb_entry_rpn [lindex $list_item 1]
         
        # Read this TLB entry from "memory"
        set addr [expr [expr $tlb_entry_num * 16] + $base]
        set tlb_tag_addr [format "0x%08x" [expr $addr + 4]]
        set tlbhi [format "0x%08x" [mem_read_word $tlb_tag_addr] ]
        set tlb_data_addr [format "0x%08x" [expr $addr + 8]]
        set tlblo [format "0x%08x" [mem_read_word $tlb_data_addr] ]

        # Print it out
        puts -nonewline "$tlb_entry_num $tlb_entry_rpn "
        set tlb_tag [decode_tlb_tag $tlbhi]
        puts -nonewline "$tlb_tag "
      
        decode_tlb_hi $tlbhi
        decode_tlb_lo $tlblo

        puts ""
        incr idx
    }
}
