###############################################################################
##
## Copyright (c) 2008 Xilinx, Inc.  All rights reserved.
##
## Xilinx, Inc.
## XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
## COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
## ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
## STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
## IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
## FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
## XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
## THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
## ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
## FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
## AND FITNESS FOR A PARTICULAR PURPOSE.
##
## Filename:    ppc440_tlb.tcl
## Author:      Brian Hill
##
## Description:
## Script to display/decode a few PPC440 registers
##
## $Id: ppc440.tcl,v 1.2 2008/05/20 21:30:47 bhill Exp $
##
###############################################################################
#
# EXAMPLE USAGE:
# XMD% ppc440_print
# PVR:  0x7ff21910 Silicon: 1.1
# MSR:  0x00000000 PRpriv
# TCR:  0x00000000
# TSR:  0xdc000000 ENW WIS
# ESR:  0x00800000 ST
# 

puts "ppc440_print"

# Print and decode contents of Processor Version Register
proc ppc440_pvr_print {} {
    set regval [string trimleft [srrd pvr] "pvr: "]
    set regval [format "0x%08x" 0x$regval ]
    puts -nonewline "PVR:  $regval "

    set pcf [expr [expr $regval >> 16] & 0xF]
    if {$pcf != 2} {
        puts "Unknown Core Family"
        return
    }

    if {$regval == 0x7ff21912} { 
        puts -nonewline "Silicon: 2.1 "
    }
    if {$regval == 0x7ff21911} { 
        puts -nonewline "Silicon: 2.0 "
    }
    if {$regval == 0x7ff21910} { 
        puts -nonewline "Silicon: 1.1 "
    }
    puts ""
}

# Print and decode contents of Machine State Register
proc ppc440_msr_print {} {
    set regval [string trimleft [rrd msr] "msr: "]
    set regval [format "0x%08x" 0x$regval ]
    puts -nonewline "MSR:  $regval "

    # WE Wait State Enable
    if {$regval & 0x00040000} {
        puts -nonewline "WE "
    }
    # CE Critical Interrupt Enable
    if {$regval & 0x00020000} {
        puts -nonewline "CE "
    }
    # EE External Interrupt Enable
    if {$regval & 0x00008000} {
        puts -nonewline "EE "
    }
    # PR Privilege Mode
    if {$regval & 0x00004000} {
        puts -nonewline "PRuser "
    } else {
        puts -nonewline "PRpriv "
    }
    # Floating Point Available
    if {$regval & 0x00002000} {
        puts -nonewline "FP "
    }
    # ME Machine Check Enable
    if {$regval & 0x00001000} {
        puts -nonewline "ME "
    }
    # FE0 Floating-point Exception mode 0
    if {$regval & 0x00000800} {
        puts -nonewline "FE0 "
    }
    # DWE Debug Wait Enable
    if {$regval & 0x00000400} {
        puts -nonewline "DWE "
    }
    # DE Debug Interrupt Enable
    if {$regval & 0x00000200} {
        puts -nonewline "DE "
    }
    # FE1 Floating-point Exception mode 1
    if {$regval & 0x00000100} {
        puts -nonewline "FE1 "
    }
    # IS Instruction address space
    if {$regval & 0x00000020} {
        puts -nonewline "IS "
    }
    # DS Data address space
    if {$regval & 0x00000010} {
        puts -nonewline "DS "
    }

    puts ""
}

# Print and decode contents of TCR Timer Control Register
proc ppc440_tcr_print {} {
    set regval [string trimleft [srrd tcr] "tcr: "]
    set regval [format "0x%08x" 0x$regval ]
    puts -nonewline "TCR:  $regval "

    # WIE Watchdog Interrupt Enable
    if {$regval & 0x08000000} {
        puts -nonewline "WIE "
    }
    # DIE Decrementer-Interrupt Enable
    if {$regval & 0x04000000} {
        puts -nonewline "DIE "
    }
    # FIT FIT-Interrupt Enable
    if {$regval & 0x00800000} {
        puts -nonewline "FIT "
    }

    puts ""
}

# Print and decode contents of TSR Timer Status Register
proc ppc440_tsr_print {} {
    set regval [string trimleft [srrd tsr] "tsr: "]
    set regval [format "0x%08x" 0x$regval ]
    puts -nonewline "TSR:  $regval "

    # ENW Enable Next Watchdog
    if {$regval & 0x80000000} {
        puts -nonewline "ENW "
    }
    # WIS Watchdog Interrupt Status
    if {$regval & 0x40000000} {
        puts -nonewline "WIS "
    }

    puts ""
}

# Exception Syndrome Register
proc ppc440_esr_print {} {
    set regval [string trimleft [srrd esr] "esr: "]
    set regval [format "0x%08x" 0x$regval ]
    puts -nonewline "ESR:  $regval "

    if {$regval & 0x80000000} {
        # Machine Check - Instruction
        puts -nonewline "MCI "
    }
    if {$regval & 0x08000000} {
        # Program - Illegal Instruction
        puts -nonewline "PIL "
    }
    if {$regval & 0x04000000} {
        # Program - Privileged Instruction
        puts -nonewline "PPR "
    }
    if {$regval & 0x02000000} {
        # Program - Trap Instruction
        puts -nonewline "PTR "
    }
    if {$regval & 0x01000000} {
        # Floating Point Exception
        puts -nonewline "FP "
    }
    if {$regval & 0x00800000} {
        # Data Storage - Store Instruction
        puts -nonewline "ST "
    }
    if {$regval & 0x00300000} {
        # Data Storage Interrupt - Locking Exception
        puts -nonewline "DLK "
    }
    if {$regval & 0x00080000} {
        # AuxProcessor Operation
        puts -nonewline "AP "
    }
    if {$regval & 0x00040000} {
        # Program Interrupt - Unimplemented Operation
        puts -nonewline "PUO "
    }
    if {$regval & 0x00020000} {
        # Byte Ordering
        puts -nonewline "BO "
    }
    if {$regval & 0x00010000} {
        # Program Interrupe - Imprecise Exception
        puts -nonewline "PIE "
    }
    if {$regval & 0x00000010} {
        # Program Interrupt - Condition Register Enable
        puts -nonewline "PCRE "
    }
    if {$regval & 0x00000008} {
        # Program Interrupt - Floating Point Compare
        puts -nonewline "PCMP "
    }
    puts ""
}

proc ppc440_print {} {
    ppc440_pvr_print
    ppc440_msr_print
    ppc440_tcr_print
    ppc440_tsr_print
    ppc440_esr_print
}
