###############################################################################
##
## Copyright (c) 2008 Xilinx, Inc.  All rights reserved.
##
## Xilinx, Inc.
## XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
## COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
## ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
## STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
## IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
## FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
## XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
## THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
## ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
## FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
## AND FITNESS FOR A PARTICULAR PURPOSE.
##
## Filename:    ppc440_dcache.tcl
## Author:      Brian Hill
##
## Description:
## Print PPC440 Data Cache contents.  XMD has pseudo-memory mapped the
## D-Cache tag and data areas.
##
## $Id: ppc440_dcache.tcl,v 1.1 2008/04/02 22:42:31 bhill Exp $
##
###############################################################################
#
# EXAMPLE USAGE:
#
# Set:     0-63
# Way:     0-15
# TRA:     Tag Real Addres (most significant 23 bits)
# Address: calculated with Tag + Set
# Drt:     which words of cache line are dirty
#
# XMD% ppc440_dcache_print 0x78008000 0x78000000
# Set Way TRA        Address  Drt Data
#  7   0  0x00024200 000242e0 0xf 00000010000000110000001200000013
#                                 00000014000000150000001600000017
# 15   0  0x00004200 000043e0 0xf 00000000000000010000000200000003
#                                 00000004000000050000000600000007
# 15   1  0x00014200 000143e0 0xf 00000008000000090000000a0000000b
#                                 0000000c0000000d0000000e0000000f
# 15   2  0x00034200 000343e0 0xf 00000020000000210000002200000023
#                                 00000024000000250000002600000027
# 
# XMD% mrd 0x242e0 8
#    242E0:   00000010
#    242E4:   00000011
#    242E8:   00000012
#    242EC:   00000013
#    242F0:   00000014
#    242F4:   00000015
#    242F8:   00000016
#    242FC:   00000017
# 

puts "ppc440_dcache_print"

# ppc440_dcache_print:
#
proc ppc440_dcache_print {tag data} {
    set line 0
    set linesize 32 

    puts "Set Way TRA        Address  Drt Data"

    # examine all 16 sets for all 64 Ways (1024 cache lines, 32K cache)
    while {$line < 1024} {
        # read the tag for this line
        set tagw0 [format "0x%08x" [mem_read_word $tag]]
        set tagw1 [format "0x%08x" [mem_read_word [expr $tag + 4]]]

        set tag_tra  [format "0x%08x" [expr $tagw0 & 0xFFFFFE00]]
        set tag_vld  [format "0x%08x" [expr [expr $tagw0 >> 7] & 1]]
        set tag_tera [format "0x%08x" [expr $tagw0 & 0xF]]

        set tag_drty [format "0x%x" [expr [expr $tagw1 >> 4] & 0xF]]

        # Is this line valid (presently in use) ?
        if {$tag_vld} {
            set cache_way [format "%2d" [expr $line / 16]]
            set cache_set [format "%2d" [expr $line % 16]]

            # Calculate the address which is cached by adding Tag Real Address
            # and the cache set
            set address [expr $tag_tra + [expr $cache_set << 5]]
            set address [format "%08x" $address]
            puts -nonewline "$cache_set  $cache_way  $tag_tra $address "
            puts -nonewline "$tag_drty "

            # Print the cache data for this line
            set i 0
            while {$i < $linesize} {
                set dataval [format "%08x" [mem_read_word [expr $data + $i]]]
                puts -nonewline "$dataval"

                # Print the data on two lines
                incr i 4
                if {$i == 16} {
                    puts -nonewline "\n                                "
                }
            }
            puts ""
        }

        # Next cache line
        incr tag  $linesize
        incr data $linesize
        incr line
    }
}

