###############################################################################
##
## Copyright (c) 2008 Xilinx, Inc.  All rights reserved.
##
## Xilinx, Inc.
## XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
## COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
## ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
## STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
## IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
## FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
## XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
## THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
## ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
## FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
## AND FITNESS FOR A PARTICULAR PURPOSE.
##
## Filename:    ppc440_icache.tcl
## Author:      Brian Hill
##
## Description:
## Print PPC440 Instruction Cache contents.  XMD has pseudo-memory mapped the
## D-Cache tag and data areas.
##
## $Id: ppc440_icache.tcl,v 1.1 2008/04/03 17:01:48 bhill Exp $
##
###############################################################################
#
# EXAMPLE USAGE:
# TEA:    Tag Effective (virtual) Address
#
# XMD% ppc440_icache_print 0x70008000 0x70000000
# Set      Way TEA        Data
#  0(0x000)  0 0xffff0800 801f00147c090214901f0010801f0010
#                         7c001764901f0008801f0008901f000c
#  1(0x020)  0 0xffff0800 7c0004ac801f000c540905a8801f0010
#                         7d2017a44c00012c801f00185400f0be
# ..... omitted ....
# 15(0x1e0)  0 0xffff0600 3800000f901f0014e5ff0850801f0018
#                         540007be2f800000e19e0838813f0014
# ..... omitted ....
# 
# standalone BSP, virtual address = physical address.
# 0xffff0600 + 0x1e0 = 0xffff07e0
#
# XMD% mrd 0xffff07e0 8
# FFFF07E0:   3800000F
# FFFF07E4:   901F0014
# FFFF07E8:   48000068
# FFFF07EC:   801F0018
# FFFF07F0:   540007BE
# FFFF07F4:   2F800000
# FFFF07F8:   419E0040
# FFFF07FC:   813F0014
# 
# powerpc-eabi-objdump -S executable.elf
# ...
# ffff07e0:     38 00 00 0f     li      r0,15
# ffff07e4:     90 1f 00 14     stw     r0,20(r31)
# ffff07e8:     48 00 00 68     b       ffff0850 <XCache_EnableICache+0xd4>
# ffff07ec:     80 1f 00 18     lwz     r0,24(r31)
# ffff07f0:     54 00 07 be     clrlwi  r0,r0,30
# ffff07f4:     2f 80 00 00     cmpwi   cr7,r0,0
# ffff07f8:     41 9e 00 40     beq-    cr7,ffff0838 <XCache_EnableICache+0xbc>
# ffff07fc:     81 3f 00 14     lwz     r9,20(r31)
#

puts "ppc440_icache_print"

# ppc440_icache_print:
#
proc ppc440_icache_print {tag data} {
    set line 0
    set linesize 32 

    puts "Set      Way TEA        Data"

    # examine all 16 sets for all 64 Ways (1024 cache lines, 32K cache)
    while {$line < 1024} {
        # read the tag for this line
        set tagw0 [format "0x%08x" [mem_read_word $tag]]
        set tagw1 [format "0x%08x" [mem_read_word [expr $tag + 4]]]

        set tag_tea  [format "0x%08x" [expr $tagw0 & 0xFFFFFE00]]
        set tag_vld  [format "0x%08x" [expr [expr $tagw0 >> 7] & 1]]

        set tag_ts   [format "0x%x" [expr [expr $tagw1 >> 9] & 0x1]]
        set tag_td   [format "0x%x" [expr [expr $tagw1 >> 8] & 0x1]]
        set tag_tid  [format "0x%x" [expr $tagw1 & 0xFF]]

        # Is this line valid (presently in use) ?
        if {$tag_vld} {
            set cache_way [format "%2d" [expr $line / 16]]
            set cache_set [format "%2d" [expr $line % 16]]
            set set_physaddr [format "0x%03x" [expr $cache_set << 5]]

            puts -nonewline "$cache_set\($set_physaddr\) $cache_way $tag_tea "

            # Print the cache data for this line
            set i 0
            while {$i < $linesize} {
                set dataval [format "%08x" [mem_read_word [expr $data + $i]]]
                puts -nonewline "$dataval"

                # Print the data on two lines
                incr i 4
                if {$i == 16} {
                    puts -nonewline "\n                        "
                }
            }
            puts ""
        }

        # Next cache line
        incr tag  $linesize
        incr data $linesize
        incr line
    }
}

