###############################################################################
##
## Copyright (c) 2008 Xilinx, Inc.  All rights reserved.
##
## Xilinx, Inc.
## XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
## COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
## ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
## STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
## IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
## FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
## XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
## THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
## ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
## FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
## AND FITNESS FOR A PARTICULAR PURPOSE.
##
## Filename:    ppc440_tlb.tcl
## Author:      Brian Hill
##
## Description:
## Script to display/decode PPC440 TLB entries via XMD "memory map" of these.
##
## $Id: ppc440_tlb.tcl,v 1.2 2008/07/01 17:29:31 bhill Exp $
##
###############################################################################
#
# EXAMPLE USAGE:
# XMD% ppc440_read_tlb 0x70020000
# #  RPN        EPN        TS SIZE  ERPN U0U1U2U3 W I M G E UX UW UR SX SW SR
# 63 0x00000000 0xc0000000    256MB 0               I   G            SX SW SR
# 57 0x00176000 0x30013000    4KB   0               I   G   UX    UR SX    SR
# 14 0x00176000 0x30013000    4KB   0               I   G   UX    UR SX    SR
# 21 0x00181000 0x10060000    4KB   0               I M G      UW UR    SW SR
# 44 0x00186000 0x7fa1b000    4KB   0               I M G      UW UR    SW SR
# 26 0x00242000 0x10012000    4KB   0               I   G   UX    UR SX    SR
# 15 0x00243000 0x0ffed000    4KB   0               I M G   UX UW UR SX SW SR
# 36 0x00246000 0x0ff1d000    4KB   0               I   G   UX    UR SX    SR
# 39 0x0024a000 0x10005000    4KB   0               I   G   UX    UR SX    SR
# ...... removed from display ....
# 59 0x03eb3000 0x0fea2000    4KB   0               I   G         UR       SR
#  1 0x03eb5000 0x0ff1e000    4KB   0               I   G   UX    UR SX    SR
# 53 0x03eb6000 0x0ff1c000    4KB   0               I   G   UX    UR SX    SR
# 45 0x03ebe000 0x0ff79000    4KB   0               I   G   UX    UR SX    SR
# 34 0x03ee1000 0x0ffeb000    4KB   0               I M G   UX    UR SX    SR
# 62 0xd0000000 0xd0000000    16KB  0               I   G               SW SR
# 

puts "ppc440_read_tlb"

# tlb_list_insert:
# Insert new_item into tlb_list by order of increasing RPN
proc tlb_list_insert {tlb_list new_item} {

    set new_tlb_entry_num [lindex $new_item 0]
    set new_tlb_entry_rpn [lindex $new_item 1]

    set tlb_list_len [llength $tlb_list]
    set item 0
    # Find where new_item belongs. Walk the entire list.
    while {$item < $tlb_list_len} {
        set list_item [lindex $tlb_list $item]
        set tlb_entry_num [lindex $list_item 0]
        set tlb_entry_rpn [lindex $list_item 1]

        # new_item RPN smaller than this list item RPN ?
        if {$new_tlb_entry_rpn < $tlb_entry_rpn} {
            # Insert new_item before the current item
            set tlb_list [linsert $tlb_list $item $new_item]
            break
        }

        # new_item RPN equal to this list item RPN ?
        # Presumably the EPN/TS/TID is different...
        if {$new_tlb_entry_rpn == $tlb_entry_rpn} {
           # Secondary sort by TLB Index

           # Find where new_item belongs. 
           while {$item < $tlb_list_len} {
               set list_item [lindex $tlb_list $item]
               set tlb_entry_num [lindex $list_item 0]
               set tlb_entry_rpn [lindex $list_item 1]

               # No longer the same RPN (new item has largest index)
               if {$new_tlb_entry_rpn != $tlb_entry_rpn} {
                   # Insert new_item before the current item
                   set tlb_list [linsert $tlb_list $item $new_item]
                   break
               }

               if {$new_tlb_entry_num < $tlb_entry_num} {
                   # Insert new_item before the current item
                   set tlb_list [linsert $tlb_list $item $new_item]
                   break
               }

               # Next list item
               incr item
           }

           # New item at the end of list.
           break
        }

        # Next list item
        incr item
    }

    # new_item at the end of list? (Largest item in the list)
    if {$item == $tlb_list_len} {
        # Insert new_item at end of list
        set tlb_list [linsert $tlb_list $item $new_item]
    }

    return $tlb_list
}


# What is the Effective Page Number
proc ppc440_tlb_pid_epn {tlb_pid} {
    set epn [format "0x%08x" [expr $tlb_pid & 0xFFFFFC00]]

    return $epn
}

# Is this TLB entry Valid?
proc ppc440_tlb_pid_v {tlb_pid} {
    return [expr $tlb_pid & [expr 1 << 9] ]
}

# Decode Page ID fields (except EPN)
proc ppc440_tlb_pid_print {tlb_pid} {
    # Translation address Space
    set ts [expr $tlb_pid & [expr 1 << 8] ]
    if {$ts} {
        puts -nonewline "TS "
    } else {
        puts -nonewline "   "
    }

    set size [expr [expr $tlb_pid >> 4] & 0xF]
    switch $size {
    0 {puts -nonewline "1KB   "}
    1 {puts -nonewline "4KB   "}
    2 {puts -nonewline "16KB  "}
    3 {puts -nonewline "64KB  "}
    4 {puts -nonewline "256KB "}
    5 {puts -nonewline "1MB   "}
    7 {puts -nonewline "16MB  "}
    9 {puts -nonewline "256MB "}
    default {puts -nonewline "Invld "}
    }
}

# What is the Real Page Number (physical address)
proc ppc440_tlb_at_rpn {tlb_at} {
    set epn [format "0x%08x" [expr $tlb_at & 0xFFFFFC00]]

    return $epn
}

# Decode Address Translation fields (except RPN)
proc ppc440_tlb_at_print {tlb_at} {
    set erpn [expr $tlb_at & 0xF]
    puts -nonewline "$erpn    "
}

# Decode Storage Attributes and Access Control fields
proc ppc440_tlb_sa_ac_print {tlb_sa} {

    if {$tlb_sa & [expr 1 << 15]} {
        puts -nonewline "U0"
    } else {
        puts -nonewline "  "
    }
    if {$tlb_sa & [expr 1 << 14]} {
        puts -nonewline "U1"
    } else {
        puts -nonewline "  "
    }
    if {$tlb_sa & [expr 1 << 13]} {
        puts -nonewline "U2"
    } else {
        puts -nonewline "  "
    }
    if {$tlb_sa & [expr 1 << 12]} {
        puts -nonewline "U3 "
    } else {
        puts -nonewline "   "
    }

    # Write-thru (cache copyback)
    if {$tlb_sa & [expr 1 << 11]} {
        puts -nonewline "W "
    } else {
        puts -nonewline "  "
    }

    # cache Inhibit
    if {$tlb_sa & [expr 1 << 10]} {
        puts -nonewline "I "
    } else {
        puts -nonewline "  "
    }

    # Memory coherence required
    if {$tlb_sa & [expr 1 << 9]} {
        puts -nonewline "M "
    } else {
        puts -nonewline "  "
    }

    # Guarded
    if {$tlb_sa & [expr 1 << 8]} {
        puts -nonewline "G "
    } else {
        puts -nonewline "  "
    }

    # Endian
    if {$tlb_sa & [expr 1 << 7]} {
        puts -nonewline "E "
    } else {
        puts -nonewline "  "
    }

    # User eXecute
    if {$tlb_sa & [expr 1 << 5]} {
        puts -nonewline "UX "
    } else {
        puts -nonewline "   "
    }

    # User Write
    if {$tlb_sa & [expr 1 << 4]} {
        puts -nonewline "UW "
    } else {
        puts -nonewline "   "
    }

    # User Read
    if {$tlb_sa & [expr 1 << 3]} {
        puts -nonewline "UR "
    } else {
        puts -nonewline "   "
    }

    # Supervisor eXecute
    if {$tlb_sa & [expr 1 << 2]} {
        puts -nonewline "SX "
    } else {
        puts -nonewline "   "
    }

    # Supervisor Write
    if {$tlb_sa & [expr 1 << 1]} {
        puts -nonewline "SW "
    } else {
        puts -nonewline "   "
    }

    # Supervisor Read
    if {$tlb_sa & [expr 1 << 0]} {
        puts -nonewline "SR "
    } else {
        puts -nonewline "   "
    }
}

proc ppc440_read_tlb {base} {
    set addr $base

    puts -nonewline "#  RPN        EPN        TS SIZE  ERPN U0U1U2U3 W I M G "
    puts "E UX UW UR SX SW SR" 

    # create an ordered list of the TLB entries, sorted by RPN
    set tlb_num 0
    while {$tlb_num < 64} {
        set tlb_pid_addr [format "0x%08x" [expr $addr + 4]]
        set tlb_at_addr  [format "0x%08x" [expr $addr + 8]]

        set tlb_pid [format "0x%08x" [mem_read_word $tlb_pid_addr] ]
        set vld [ppc440_tlb_pid_v $tlb_pid]

        # Valid entry?
        if {$vld} {
            set tlb_at [format "0x%08x" [mem_read_word $tlb_at_addr] ]
            set tlb_rpn [list $tlb_num [ppc440_tlb_at_rpn $tlb_at]]
            # First one?
            if { [info exists tlb_list] } {
                # Insert this entry into the list
                set tlb_list [tlb_list_insert $tlb_list $tlb_rpn]
            } else {
                # First list item
                set tlb_list [list $tlb_rpn]
            }
        }

        incr addr 16
        set addr [format "0x%08x" $addr]
        incr tlb_num
    }

    # Print all valid TLB entries as present in the ordered list
    set tlb_list_len [llength $tlb_list]
    set idx 0
    while {$idx < $tlb_list_len} {
        # Get TLB entry from the list
        set list_item [lindex $tlb_list $idx]
        # Fetch TLB entry number (0-63) and that entry's RPN from this item
        set tlb_entry_num [format "%2d" [lindex $list_item 0]]
        set tlb_entry_rpn [lindex $list_item 1]

        # Read this TLB entry from "memory"
        set addr [expr [expr $tlb_entry_num * 16] + $base]
        set tlb_pid_addr [format "0x%08x" [expr $addr + 4]]
        set tlb_at_addr  [format "0x%08x" [expr $addr + 8]]
        set tlb_sa_addr  [format "0x%08x" [expr $addr + 12]]

        set tlb_pid [format "0x%08x" [mem_read_word $tlb_pid_addr] ]
        set epn [ppc440_tlb_pid_epn $tlb_pid]
        set vld [ppc440_tlb_pid_v $tlb_pid]

        set tlb_at [format "0x%08x" [mem_read_word $tlb_at_addr] ]
        set tlb_sa [format "0x%08x" [mem_read_word $tlb_sa_addr] ]

        puts -nonewline "$tlb_entry_num $tlb_entry_rpn $epn "
        ppc440_tlb_pid_print $tlb_pid
        ppc440_tlb_at_print $tlb_at
        ppc440_tlb_sa_ac_print $tlb_sa
        puts ""

        incr idx
    }
}
