###############################################################################
##
## Copyright (c) 2008 Xilinx, Inc.  All rights reserved.
##
## Xilinx, Inc.
## XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
## COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
## ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
## STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
## IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
## FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
## XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
## THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
## ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
## FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
## AND FITNESS FOR A PARTICULAR PURPOSE.
##
## Filename:    ppc_bt.tcl
## Author:      Brian Hill
##
## Description:
## Prints a "Back Trace" from wherever the machine is presently at.
## The last downloaded ELF file is used with addr2line to include symbolic
## information.  The exeucable can only be found if the user has not changed
## directories within XMD.
##
## $Id: ppc_bt.tcl,v 1.4 2008/06/24 22:12:25 bhill Exp $
##
###############################################################################
#
# Example Usage
#
# XMD% ppc_bt
# PC:   0xeee00060            IVOR6             xvectors.S:511
# LR:   0x00000284            gen_illegal_instr TestApp_Crash.c:97
# R1:   0x00003fb0 0x00000278 gen_illegal_instr TestApp_Crash.c:95
# R1:   0x00003fd0 0x00000778 choose_crash      TestApp_Crash.c:285
# R1:   0x00003ff8 0x0000082c main              TestApp_Crash.c:319
# R1:   0x00004008 0x000000e4 fpu_init_done     xil-crt0.S:0
# XMD%

puts "ppc_bt"

# ppc_lookupsym
# Use powerpc-eabi-addr2line to find info about the symol at $addr
proc ppc_lookupsym {addr} {
    # last downloaded executable
    global elf_file

    # nothing downloaded yet?
    if {! [info exists elf_file]} {
        puts ""
        return;
    }
    # File accessable?
    if {! [file exists $elf_file]} {
        puts ""
        return;
    }

    # exec powerpc-eabi-addr2line
    # -It must be in the path.
    # -The user can not have changed directories since downloading.
    set f1 [open "|powerpc-eabi-addr2line -f -s -e $elf_file $addr" r]
    gets $f1 func
    gets $f1 file
    close $f1

    # Display function and file name for this address
    puts [format "%-17s %s" $func $file]

    return $func
}

# ppc_bt
# Display a PowerPC "Backtrace"
proc ppc_bt { {max_depth 7} } {

    # PC
    set pcval [string trimleft [srrd pc] "pc: "]
    set pcval [format "0x%08x" 0x0$pcval ]
    puts -nonewline "PC:   $pcval            "
    ppc_lookupsym $pcval

    # lr (link register)
    set lrval [string trimleft [srrd lr] "lr: "]
    set lrval [format "0x%08x" 0x0$lrval ]
    puts -nonewline "LR:   $lrval            "
    ppc_lookupsym $lrval

    # R1 (stack pointer)
    set r1val [string trimleft [rrd r1] "r1: "]
    set r1val [format "0x%08x" 0x0$r1val ]
    set i 0
    # Walk the stack
    while {($r1val != 0) && ($r1val & 0x3) == 0 && ($i < $max_depth)} {

        puts -nonewline "R1:   $r1val "

        # LR saved in this stack frame
        set saved_lr_addr [expr $r1val + 4]
        set saved_lr [format "0x%08x" [mem_read_word $saved_lr_addr] ]
        puts -nonewline "$saved_lr "
        set func [ppc_lookupsym $saved_lr]
     
        # Does this _appear_ to be an interrupt vector frame 
        # (does the function name start with _vector) ?
        # If so, fetch the saved PC/LR off the stack and look it up.
        # See ppc440_0/libsrc/standalone_v2_00_a/src/xvectors.S
        if {[regexp {_vector([0-f]+)} $func text1 vectnum]} {
           # PC saved in the exception stack frame
           set saved_pc_addr [expr $r1val + 12]
           set saved_pc [format "0x%08x" [mem_read_word $saved_pc_addr] ]
           puts -nonewline "R1*:  $r1val "
           puts -nonewline "$saved_pc "
           ppc_lookupsym $saved_pc

           # LR saved in the exception stack frame
           set saved_lr_addr [expr $r1val + 16]
           set saved_lr [format "0x%08x" [mem_read_word $saved_lr_addr] ]
           puts -nonewline "R1*:  $r1val "
           puts -nonewline "$saved_lr "
           set func [ppc_lookupsym $saved_lr]
        }

        incr i
        # R1 (stack pointer) -- next frame
        set r1val [format "0x%08x" [mem_read_word $r1val] ]
    }
}

