###############################################################################
##
## Copyright (c) 2008 Xilinx, Inc.  All rights reserved.
##
## Xilinx, Inc.
## XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
## COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
## ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
## STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
## IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
## FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
## XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
## THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
## ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
## FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
## AND FITNESS FOR A PARTICULAR PURPOSE.
##
## Filename:    ppc_dis.tcl
## Author:      Brian Hill
##
## Description:
## Prints a disassembly of instructions at the specified address.
## The powerpc-eabi-objcopy and powerpc-eabi-objdump executables are used
## to produce the disassembly -- the must be in the PATH for this script to
## operate properly.
##
## $Id: ppc_dis.tcl,v 1.1 2008/06/26 21:33:14 bhill Exp $
##
###############################################################################
#
# EXAMPLE USAGE:
# 
# XMD% ppc_dis 0xf10 8
#  f10:   94 21 ff e0     stwu    r1,-32(r1)
#  f14:   7c 08 02 a6     mflr    r0
#  f18:   93 e1 00 1c     stw     r31,28(r1)
#  f1c:   90 01 00 24     stw     r0,36(r1)
#  f20:   7c 3f 0b 78     mr      r31,r1
#  f24:   3d 20 00 00     lis     r9,0
#  f28:   38 69 63 a8     addi    r3,r9,25512
#  f2c:   48 00 0d c5     bl      1cf0 <_binary_memory_bin_end+0xdb8>
# 
# As compared to:
# powerpc-eabi-objdump -d foo.elf
#
# 00000f10 <gen_illegal_instr>:
#      f10:       94 21 ff e0     stwu    r1,-32(r1)
#      f14:       7c 08 02 a6     mflr    r0
#      f18:       93 e1 00 1c     stw     r31,28(r1)
#      f1c:       90 01 00 24     stw     r0,36(r1)
#      f20:       7c 3f 0b 78     mr      r31,r1
#      f24:       3d 20 00 00     lis     r9,0
#      f28:       38 69 63 a8     addi    r3,r9,25512
#      f2c:       48 00 0d c5     bl      1cf0 <xil_printf>

puts "ppc_dis"

# hex_to_bin:
# Convert HEX string to binary data
proc hex_to_bin {hex} {
  set binval [binary format "H*" $hex]
  return $binval
}

# memsave:
# save $words of memoray at $addr to binary $file_out
proc memsave {addr words file_out} {

    set f [open $file_out w]
    fconfigure $f \
        -translation binary \
        -encoding binary

    set word 0
    while {$word < $words} {
        set memval [format "%08x" [mem_read_word $addr]]
        set binval [hex_to_bin $memval]
        puts -nonewline $f $binval

        incr addr 4
        incr word
    }

    close $f
}

# ppc_dis:
# Disassemble $words of PPC instructions starting at $startaddr
proc ppc_dis {startaddr words} {

    # Save specified memory to a binary file
    memsave $startaddr $words "memory.bin"

    # Convert binary file to PPC object file
    exec powerpc-eabi-objcopy -I binary -O elf32-powerpc -B powerpc \
         --rename-section .data=.text \
         --change-section-address .data=$startaddr memory.bin memory.o

    # Disassemble PPC object file
    set f1 [open "|powerpc-eabi-objdump -D memory.o" r]

    # Skip initial crud
    while {[eof $f1] == 0} {
        gets $f1 line
        if {[regexp {_binary_memory_bin_start} $line text1 ]} {
            break
        }
    }

    # Display disassembly
    while {[eof $f1] == 0} {
        gets $f1 line
        puts "$line"
    }

    close $f1
    
    # Clean up temporary files
    # The TCL "file delete" procedure does not seem to operate as expected 
    exec rm memory.o
    exec rm memory.bin
}
