###############################################################################
##
## Copyright (c) 2007 Xilinx, Inc.  All rights reserved.
##
## Xilinx, Inc.
## XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
## COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
## ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
## STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
## IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
## FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
## XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
## THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
## ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
## FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
## AND FITNESS FOR A PARTICULAR PURPOSE.
##
## Filename:    xiic.tcl
## Author:      Brian Hill
##
## Description:
## Print XPS IIC controller registers. Derived from xiic_l.c
##
## $Id: xiic.tcl,v 1.3 2008/04/21 20:48:25 bhill Exp $
##
###############################################################################
#
# EXAMPLE USAGE:
# Write 0xAB to byte 131 in the EEPROM at address 0x50 
# XMD% xiic_write_byte 0x81600000 0x50 131 0xab
#
# Read byte 131 from the EEPROM at address 0x50 
# XMD% xiic_read_byte 0x81600000 0x50 131
# 0xab
#
# XMD% xiic_print 0x81600000
# 0x8160001c GIER  0x00
# 0x81600020 ISR   0xda TX_ERROR RX_FULL BNB NAAS TX_HALF
# 0x81600028 IER   0x00
# 0x81600040 RESET 0x00
# 0x81600100 CR    0x00
# 0x81600104 SR    0xc0 RX_FIFO_EMPTY TX_FIFO_EMPTY
# 0x81600108 DTR   0xa1
# 0x8160010c DRR   0x01
# 0x81600110 ADR   0x00
# 0x81600114 TFO   0x00
# 0x81600118 RFO   0x00
# 0x8160011c TBA   0x00
# 0x81600120 RFD   0x00
# 0x81600124 GPO   0x00
# 

puts "xiic_print"
puts "xiic_read_byte"
puts "xiic_write_byte"

# XIic_mClearIisr:
# Clear the specified interrupt in the Interrupt status
# register. Clearing an interrupt acknowledges it.
#
proc XIic_mClearIisr {BaseAddress InterruptMask} {
    # Read register offset definitions
    global env
    source "$env(XMD_USER_TCLDIR)/include/xiic_l.tcl"

    set addr [format "0x%08x" [expr $BaseAddress + $XIIC_IISR_OFFSET]]
    set iisr [format "0x%08x" [mem_read_word $addr] ]
    set iisr [expr $iisr & $InterruptMask]

    mwr [expr $BaseAddress + $XIIC_IISR_OFFSET] $iisr
}

# XIic_mSend7BitAddress:
# Sends the 7 bit device address during both read and write operations. 
#
proc XIic_mSend7BitAddress {BaseAddress SlaveAddress Operation} {
    # Read register offset definitions
    global env
    source "$env(XMD_USER_TCLDIR)/include/xiic_l.tcl"

    set LocalAddr [expr $SlaveAddress << 1]
    set LocalAddr [expr [expr $LocalAddr & 0xFE] + $Operation]
    mwr [expr $BaseAddress + $XIIC_DTR_REG_OFFSET] $LocalAddr
}

# iic_SendData:
# Send the specified data (byte) to the device that has been previously
# addressed on the IIC bus.  This procedure assumes that the 7 bit address
# has been sent and it should wait for the transmit of the address to complete.
#
proc iic_SendData {BaseAddress Data} {
    # Read register offset definitions
    global env
    source "$env(XMD_USER_TCLDIR)/include/xiic_l.tcl"

    set ByteCount [llength $Data]
    set ByteIndex 0
    while {$ByteCount > 0} {

        # Wait for the transmit to be empty before sending any more data
        # by polling the interrupt status register
        #
        set poll_counter 0
        while {1} {
            set addr [format "0x%08x" [expr $BaseAddress + $XIIC_IISR_OFFSET]]
            set IntrStatus [format "0x%08x" [mem_read_word $addr] ]
    
            if {$IntrStatus & [expr $XIIC_INTR_TX_ERROR_MASK + \
                                    $XIIC_INTR_ARB_LOST_MASK + \
                                    $XIIC_INTR_BNB_MASK]} {
                puts "SendData: ERROR - IISR $IntrStatus"
                return 
            }

            if {$IntrStatus & $XIIC_INTR_TX_EMPTY_MASK} {
                break;
            }

            incr poll_counter
            if {$poll_counter > 100} {
                puts "TIMEOUT: SendData TX_EMPTY"
                return
            }
        }

        # If there is more than one byte to send then put the next byte to send
        # into the transmit FIFO
        #
        if {$ByteCount > 1} {
            mwr [expr $BaseAddress + $XIIC_DTR_REG_OFFSET] \
                [lindex $Data $ByteIndex]
            incr ByteIndex
        } else {

            # Set the stop Option before sending the last byte
            # of data so that the stop Option will be generated
            # immediately following the data. This is done by clearing
            # the MSMS bit in the control register.
            mwr [expr $BaseAddress + $XIIC_CR_REG_OFFSET] \
                [expr $XIIC_CR_ENABLE_DEVICE_MASK + $XIIC_CR_DIR_IS_TX_MASK]

            # Put the last byte to send in the transmit FIFO
            mwr [expr $BaseAddress + $XIIC_DTR_REG_OFFSET] \
                [lindex $Data $ByteIndex]
            incr ByteIndex
        } 

        # Clear the latched interrupt status register and this must be done
        # after the transmit FIFO has been written to or it won't clear
        #
        XIic_mClearIisr $BaseAddress $XIIC_INTR_TX_EMPTY_MASK

        set ByteCount [expr $ByteCount - 1]
    }

    # Wait for the bus to transition to not busy before returning, the IIC
    # device cannot be disabled until this occurs.
    # Note that this is different from a receive operation because the stop
    # Option causes the bus to go not busy.
    #
    set poll_counter 0
    set addr [format "0x%08x" [expr $BaseAddress + $XIIC_IISR_OFFSET]]
    while {1} {
        set IntrStatus [format "0x%08x" [mem_read_word $addr] ]
        if {$IntrStatus & $XIIC_INTR_BNB_MASK} {
            break;
        }

        incr poll_counter
        if {$poll_counter > 100} {
            puts "TIMOUT: SendData INTR_BNB"
            return
        }
    }
}

# iic_RecvData:
# Receive a byte from the device that has been previously addressed
# on the IIC bus.  This function assumes that the 7 bit address has been sent
# and it should wait for the transmit of the address to complete.
#
proc iic_RecvData {BaseAddress} {
    # Read register offset definitions
    global env
    source "$env(XMD_USER_TCLDIR)/include/xiic_l.tcl"

    # Setup the mask to use for checking errors because when receiving one
    # byte OR the last byte of a multibyte message an error naturally
    # occurs when the no ack is done to tell the slave the last byte
    #
    set IntrStatusMask [expr $XIIC_INTR_ARB_LOST_MASK + $XIIC_INTR_BNB_MASK]

    # Wait for the previous transmit and the 1st receive to complete
    # by checking the interrupt status register of the IPIF
    #
    set poll_counter 0
    while {1} {
        set addr [format "0x%08x" [expr $BaseAddress + $XIIC_IISR_OFFSET]]
        set IntrStatus [format "0x%08x" [mem_read_word $addr] ]
        if {$IntrStatus & $XIIC_INTR_RX_FULL_MASK} {
            break;
        }
        # Check the transmit error after the receive full because when
        # sending only one byte transmit error will occur because of the
        # no ack to indicate the end of the data 
        #
        if {$IntrStatus & $IntrStatusMask} {
            puts "ERROR: RecvData $IntrStatus"
            return 0
        }

        incr poll_counter
        if {$poll_counter > 100} {
            puts "TIMEOUT: RecvData RX_FULL"
            return 0
        }
    }

    mwr [expr $BaseAddress + $XIIC_CR_REG_OFFSET] $XIIC_CR_ENABLE_DEVICE_MASK

    # Read in data from the FIFO and unthrottle the bus such that the
    # next byte is read from the IIC bus
    #
    set addr [format "0x%08x" [expr $BaseAddress + $XIIC_DRR_REG_OFFSET]]
    set ReadData [format "0x%02x" [mem_read_word $addr] ]

    return $ReadData
}

# XIic_Recv:
# Receive data as a master on the IIC bus.  This procedure receives the data
# using polled I/O and blocks until the data has been received. It only
# supports 7 bit addressing mode of operation. 
#
proc XIic_Recv {BaseAddress Address} {
    # Read register offset definitions
    global env
    source "$env(XMD_USER_TCLDIR)/include/xiic_l.tcl"

    # Tx error is enabled incase the address has no device to answer
    # with Ack. When only one byte of data, must set NO ACK before address goes
    # out therefore Tx error must not be enabled as it will go off immediately
    # and the Rx full interrupt will be checked.  If full, then the one byte
    # was received and the Tx error will be disabled without sending an error
    # callback msg.
    #
    XIic_mClearIisr $BaseAddress [expr $XIIC_INTR_RX_FULL_MASK +  \
                                       $XIIC_INTR_TX_ERROR_MASK + \
                                       $XIIC_INTR_ARB_LOST_MASK]

    # Set receive FIFO occupancy depth for 1 byte (zero based)
    #
    mwr [expr $BaseAddress + $XIIC_RFD_REG_OFFSET] 0

    # Check to see if already Master on the Bus.
    # If Repeated Start bit is not set send Start bit by setting MSMS bit else
    # Send the address.
    #
    set addr [format "0x%08x" [expr $BaseAddress + $XIIC_CR_REG_OFFSET]]
    set CntrlReg [format "0x%08x" [mem_read_word $addr] ]
    if {[expr $CntrlReg & $XIIC_CR_REPEATED_START_MASK] == 0} {
        # 7 bit slave address, send the address for a read operation
        # and set the state to indicate the address has been sent
        #
        XIic_mSend7BitAddress $BaseAddress $Address $XIIC_READ_OPERATION

        # MSMS gets set after putting data in FIFO. Start the master receive 
        # operation by setting CR Bits MSMS to Master, if the buffer is only one
        # byte, then it should not be acknowledged to indicate the end of data 
        #
        set CntrlReg [expr $XIIC_CR_MSMS_MASK + \
                           $XIIC_CR_ENABLE_DEVICE_MASK + \
                           $XIIC_CR_NO_ACK_MASK]

        # Write out the control register to start receiving data and call the
        # function to receive the data
        #
        mwr [expr $BaseAddress + $XIIC_CR_REG_OFFSET] $CntrlReg

        # Clear the latched interrupt status for the bus not busy bit which must
        # be done while the bus is busy
        #
        set poll_counter 0
        set addr [format "0x%08x" [expr $BaseAddress + $XIIC_SR_REG_OFFSET]]
        set StatusReg [format "0x%08x" [mem_read_word $addr] ]
        while {[expr $StatusReg & $XIIC_SR_BUS_BUSY_MASK] == 0} {
            set StatusReg [format "0x%08x" [mem_read_word $addr] ]

            incr poll_counter
            if {$poll_counter > 100} {
                puts "TIMEOUT: XIic_Recv SR_BUS_BUSY"
                return 0
            }
        }

        XIic_mClearIisr $BaseAddress $XIIC_INTR_BNB_MASK
    } else {
        # Already owns the Bus indicating that its a Repeated Start call.
        # 7 bit slave address, send the address for a read operation
        # and set the state to indicate the address has been sent
        #
        XIic_mSend7BitAddress $BaseAddress $Address $XIIC_READ_OPERATION
    }

    # Try to receive the data from the IIC bus 
    set ReadData [iic_RecvData $BaseAddress]

    # The receive is complete, disable the IIC device 
    #
    mwr [expr $BaseAddress + $XIIC_CR_REG_OFFSET] 0

    return $ReadData
}

# XIic_Send:
# Send data as a master on the IIC bus.  This procedure sends the data
# using polled I/O and blocks until the data has been sent.  It only supports
# 7 bit addressing mode of operation.
#
proc XIic_Send {BaseAddress iic_address write_data} {
    # Read register offset definitions
    global env
    source "$env(XMD_USER_TCLDIR)/include/xiic_l.tcl"

    # Put the address into the FIFO to be sent and indicate that the
    # operation to be performed on the bus is a write operation
    #
    XIic_mSend7BitAddress $BaseAddress $iic_address $XIIC_WRITE_OPERATION

    # Clear the latched interrupt status so that it will be updated with
    # the new state when it changes, this must be done after the address
    # is put in the FIFO
    #
    XIic_mClearIisr $BaseAddress [expr $XIIC_INTR_TX_EMPTY_MASK + \
                                       $XIIC_INTR_TX_ERROR_MASK + \
                                       $XIIC_INTR_ARB_LOST_MASK]

    # MSMS must be set after putting data into transmit FIFO, indicate the
    # direction is transmit, this device is master and enable the IIC
    # device
    #
    mwr [expr $BaseAddress + $XIIC_CR_REG_OFFSET] \
        [expr $XIIC_CR_MSMS_MASK + \
              $XIIC_CR_DIR_IS_TX_MASK + \
              $XIIC_CR_ENABLE_DEVICE_MASK]

    # Clear the latched interrupt status for the bus not busy bit which
    # must be done while the bus is busy
    #
    set poll_counter 0
    set addr [format "0x%08x" [expr $BaseAddress + $XIIC_SR_REG_OFFSET]]
    set StatusReg [format "0x%08x" [mem_read_word $addr] ]
    while {[expr $StatusReg & $XIIC_SR_BUS_BUSY_MASK] == 0} {
        set StatusReg [format "0x%08x" [mem_read_word $addr] ]

        incr poll_counter
        if {$poll_counter > 100} {
            puts "TIMEOUT: XIic_Send SR_BUS_BUSY"
            return
        }
    }
    XIic_mClearIisr $BaseAddress $XIIC_INTR_BNB_MASK

    # Send the specified data to the device on the IIC bus specified by the
    # the address
    #
    iic_SendData $BaseAddress $write_data

    # The Transmission is completed, disable the IIC device 
    #
    set addr [format "0x%08x" [expr $BaseAddress + $XIIC_CR_REG_OFFSET]]
    set ControlReg [format "0x%08x" [mem_read_word $addr] ]
    if {[expr $ControlReg & $XIIC_CR_MSMS_MASK] != 0} {
        mwr $addr [expr $ControlReg & ~$XIIC_CR_MSMS_MASK]

        set poll_counter 0
        set addr [format "0x%08x" [expr $BaseAddress + $XIIC_SR_REG_OFFSET]]
        set StatusReg [format "0x%08x" [mem_read_word $addr] ]
        while {[expr $StatusReg & $XIIC_SR_BUS_BUSY_MASK] != 0} {
            set StatusReg [format "0x%08x" [mem_read_word $addr] ]

            incr poll_counter
            if {$poll_counter > 100} {
                puts "TIMEOUT: XIic_Send SR_BUS_BUSY"
                return
            }
        }
    }

    mwr [expr $BaseAddress + $XIIC_CR_REG_OFFSET] 0
}

# xiic_read_byte:
# Read the byte at SlaveOffset from the eeprom at 7-bit SlaveAddress
#
proc xiic_read_byte {BaseAddress SlaveAddress SlaveOffset} {
    # Read register offset definitions
    global env
    source "$env(XMD_USER_TCLDIR)/include/xiic_l.tcl"

    # wait for the controller to complete previous operation
    set poll_counter 0
    set addr [format "0x%08x" [expr $BaseAddress + $XIIC_SR_REG_OFFSET]]
    set StatusReg [format "0x%08x" [mem_read_word $addr] ]
    while {[expr $StatusReg & $XIIC_SR_BUS_BUSY_MASK]} {
        set StatusReg [format "0x%08x" [mem_read_word $addr] ]

        incr poll_counter
        if {$poll_counter > 100} {
            puts "TIMEOUT: xiic_read_byte SR_BUS_BUSY"
            return
        }
    }
    #
    # Set the address register to the specified address by writing
    # the address to the device
    #
    XIic_Send $BaseAddress $SlaveAddress $SlaveOffset

    #
    # Read the byte at the specified address from the EEPROM
    #
    set ReadData [XIic_Recv $BaseAddress $SlaveAddress]

    return $ReadData
}

# xiic_write_byte:
# Write supplied byte at SlaveOffset from the eeprom at 7-bit SlaveAddress
#
proc xiic_write_byte {BaseAddress SlaveAddress SlaveOffset WriteData} {
    # Read register offset definitions
    global env
    source "$env(XMD_USER_TCLDIR)/include/xiic_l.tcl"

    # wait for the controller to complete previous operation
    set poll_counter 0
    set addr [format "0x%08x" [expr $BaseAddress + $XIIC_SR_REG_OFFSET]]
    set StatusReg [format "0x%08x" [mem_read_word $addr] ]
    while {[expr $StatusReg & $XIIC_SR_BUS_BUSY_MASK]} {
        set StatusReg [format "0x%08x" [mem_read_word $addr] ]

        incr poll_counter
        if {$poll_counter > 100} {
            puts "TIMEOUT: xiic_read_byte SR_BUS_BUSY"
            return
        }
    }

    #
    # Write the byte at the specified address from the EEPROM
    #
    XIic_Send $BaseAddress $SlaveAddress [list $SlaveOffset $WriteData]
}

proc xiic_print {base} {
    # Read register offset definitions
    global env
    source "$env(XMD_USER_TCLDIR)/include/xiic_l.tcl"

    set addr [format "0x%08x" [expr $base + $XIIC_DGIER_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts "$addr GIER  $mval "

    set addr [format "0x%08x" [expr $base + $XIIC_IISR_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts -nonewline "$addr ISR   $mval "
    if {$mval & $XIIC_INTR_ARB_LOST_MASK} {
        puts -nonewline "ARB_LOST "
    }
    if {$mval & $XIIC_INTR_TX_ERROR_MASK} {
        puts -nonewline "TX_ERROR "
    }
    if {$mval & $XIIC_INTR_TX_EMPTY_MASK} {
        puts -nonewline "TX_EMPTY "
    }
    if {$mval & $XIIC_INTR_RX_FULL_MASK} {
        puts -nonewline "RX_FULL "
    }
    if {$mval & $XIIC_INTR_BNB_MASK} {
        puts -nonewline "BNB "
    }
    if {$mval & $XIIC_INTR_AAS_MASK} {
        puts -nonewline "AAS "
    }
    if {$mval & $XIIC_INTR_NAAS_MASK} {
        puts -nonewline "NAAS "
    }
    if {$mval & $XIIC_INTR_TX_HALF_MASK} {
        puts -nonewline "TX_HALF "
    }
    puts ""

    set addr [format "0x%08x" [expr $base + $XIIC_IIER_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts "$addr IER   $mval "

    set addr [format "0x%08x" [expr $base + $XIIC_RESETR_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts "$addr RESET $mval "

    set addr [format "0x%08x" [expr $base + $XIIC_CR_REG_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts -nonewline "$addr CR    $mval "
    if {$mval & $XIIC_CR_ENABLE_DEVICE_MASK} {
        puts -nonewline "ENABLE "
    }
    if {$mval & $XIIC_CR_TX_FIFO_RESET_MASK} {
        puts -nonewline "TX_FIFO_RESET "
    }
    if {$mval & $XIIC_CR_MSMS_MASK} {
        puts -nonewline "MSMS "
    }
    if {$mval & $XIIC_CR_DIR_IS_TX_MASK} {
        puts -nonewline "DIR_IS_TX "
    }
    if {$mval & $XIIC_CR_NO_ACK_MASK} {
        puts -nonewline "NO_ACK "
    }
    if {$mval & $XIIC_CR_REPEATED_START_MASK} {
        puts -nonewline "REPEATED_START "
    }
    if {$mval & $XIIC_CR_GENERAL_CALL_MASK} {
        puts -nonewline "GENERAL_CALL "
    }
    puts ""

    set addr [format "0x%08x" [expr $base + $XIIC_SR_REG_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts -nonewline "$addr SR    $mval "
    if {$mval & $XIIC_SR_GEN_CALL_MASK} {
        puts -nonewline "GEN_CALL "
    }
    if {$mval & $XIIC_SR_ADDR_AS_SLAVE_MASK} {
        puts -nonewline "ADDR_AS_SLAVE "
    }
    if {$mval & $XIIC_SR_BUS_BUSY_MASK} {
        puts -nonewline "BUS_BUSY "
    }
    if {$mval & $XIIC_SR_MSTR_RDING_SLAVE_MASK} {
        puts -nonewline "MSTR_RDING_SLAVE "
    }
    if {$mval & $XIIC_SR_TX_FIFO_FULL_MASK} {
        puts -nonewline "TX_FIFO_FULL "
    }
    if {$mval & $XIIC_SR_RX_FIFO_FULL_MASK} {
        puts -nonewline "RX_FIFO_FULL "
    }
    if {$mval & $XIIC_SR_RX_FIFO_EMPTY_MASK} {
        puts -nonewline "RX_FIFO_EMPTY "
    }
    if {$mval & $XIIC_SR_TX_FIFO_EMPTY_MASK} {
        puts -nonewline "TX_FIFO_EMPTY "
    }
    puts ""

    set addr [format "0x%08x" [expr $base + $XIIC_DTR_REG_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts "$addr DTR   $mval"

    set addr [format "0x%08x" [expr $base + $XIIC_DRR_REG_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts "$addr DRR   $mval"

    set addr [format "0x%08x" [expr $base + $XIIC_ADR_REG_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts "$addr ADR   $mval"

    set addr [format "0x%08x" [expr $base + $XIIC_TFO_REG_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts "$addr TFO   $mval"

    set addr [format "0x%08x" [expr $base + $XIIC_RFO_REG_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts "$addr RFO   $mval"

    set addr [format "0x%08x" [expr $base + $XIIC_TBA_REG_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts "$addr TBA   $mval"

    set addr [format "0x%08x" [expr $base + $XIIC_RFD_REG_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts "$addr RFD   $mval"

    set addr [format "0x%08x" [expr $base + $XIIC_GPO_REG_OFFSET]]
    set mval [format "0x%02x" [mem_read_word $addr] ]
    puts "$addr GPO   $mval"
}
